﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ScrabbleClient.ScrabbleService;
using System.ServiceModel;
using System.ServiceModel.Security;
using ServerLibrary.WCF.DTO;
using System.Threading;

namespace ScrabbleClient.Domain
{
    public sealed class ClientManager
    {
        //La classe ClientManager serà un singleton que gestionarà el control del client
        //Totes les finestres podran accedir a la classe facilment.
        //Encapsularà les crides al proxy del servei i realitzara un primer control basic d'errors

        /*
         * Gestió del singleton
         */
        private static readonly ClientManager instance = new ClientManager();

        public static ClientManager Instance
        {
            get
            {
                return instance;
            }
        }

        private ClientManager() { }

        /*
         * Gestio de observers (Patró observador)
         */
        private List<IGameObserver> observers = new List<IGameObserver>();


        public void AddObserver(IGameObserver observer)
        {
            observers.Add(observer);
        }

        public void RemoveObserver(IGameObserver observer)
        {
            observers.Remove(observer);
        }

        public void YourTurn(GameInfoDTO gameInfo, List<List<BoxDTO>> boardInfo, List<TileDTO> rackInfo, MoveInfoDTO moveInfo)
        {
            foreach (IGameObserver observer in observers)
            {
                observer.YourTurn(gameInfo, boardInfo, rackInfo, moveInfo);
            }
        }

        public void OpponentResigned(GameInfoDTO gameInfo)
        {
            foreach (IGameObserver observer in observers)
            {
                observer.OpponentResigned(gameInfo);
            }
        }

        /*
         * Comunicació amb el servei
         */
        private ScrabbleServiceClient scrabbleProxy;

        Thread hbThread = null;

        private UserDTO userInfo;

        public UserDTO User { get { return userInfo; } }

        public void HeartBeat()
        {
            int i;
            while (true)
            {
                try
                {                   
                    /*Dormim el thread 30 segons 
                     * Ho fem amb 30 sleeps de 1 segon, en lloc de un sleep de 30 segons, ja que d'aquesta forma, si durant l'espera
                     * l'usuari tanca sessió nomes haurem d'esperar 1 segon a que finalitzi aquest thread
                     * */
                    i = 0;
                    while (i < 30)
                    {
                        Thread.Sleep(1000);
                        i++;
                    }

                    scrabbleProxy.HeartBeat();
                }
                catch (ThreadAbortException)
                {
                    //No fem res, es l'excepcio que es llança per parar el thread
                    throw;
                }
                catch (MessageSecurityException ex)
                {
                    scrabbleProxy.Abort();
                    throw new Exception(ex.InnerException.Message);
                }
                catch (Exception)
                {
                    
                }
                    
            }
        }

        public void CreateUser(string name, string password)
        {
            try
            {
                var context = new InstanceContext(new ServiceCallback());
                scrabbleProxy = new ScrabbleServiceClient(context);

                //Assignem credencials especials perque ens autentiqui i ens permeti crear el nou usuari
                scrabbleProxy.ClientCredentials.UserName.UserName = "CreateNewUser";
                scrabbleProxy.ClientCredentials.UserName.Password = "{35CCFA7A-96D7-456A-A927-B67012EF13A7}";

                scrabbleProxy.CreateUser(new UserDTO() { Name = name, Password = password });

                Login(name, password);
            }
            catch (FaultException ex)
            {
                scrabbleProxy.Abort();
                throw new Exception(ex.Message);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                throw new Exception(ex.InnerException.Message);
            }
        }

        private void OpenSession(string name, string password)
        {
            var context = new InstanceContext(new ServiceCallback());
            scrabbleProxy = new ScrabbleServiceClient(context);

            //Assignem credencials perque ens autentiqui i ens permeti identificar l'usuari
            scrabbleProxy.ClientCredentials.UserName.UserName = name;
            scrabbleProxy.ClientCredentials.UserName.Password = password;

            //Si el login es correcte ens torna un objecte amb les dades de l'usuari, que desarem per futures referències
            userInfo = scrabbleProxy.LogIn();
            userInfo.Password = password;

            //Creem un thread que faci un heartbeat periodic al server
            hbThread = new Thread(new ThreadStart(HeartBeat));
            hbThread.Start();
        }

        public void Login(string name, string password)
        {
            try
            {
                OpenSession(name, password);
            }
            catch (FaultException ex)
            {
                scrabbleProxy.Abort();
                throw new Exception(ex.Message);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void Logout()
        {
            //Finalitzar el thread que fa heartbeat
            if (hbThread != null)
            {
                try
                {
                    hbThread.Abort();
                }
                catch (Exception)
                {
                }
                hbThread.Join();
            }
            hbThread = null;

            if (scrabbleProxy != null && scrabbleProxy.State == CommunicationState.Opened)
            {
                scrabbleProxy.LogOut();                     
            }
            scrabbleProxy = null; 

            userInfo = null;
            observers.Clear();
        }

        private void CheckConnectionStatus()
        {
            //Si la comunicació ha caigut, intentem tornar a fer login de forma transparent a l'usuari, si no es possible llençem una excepció
            if (scrabbleProxy.State != CommunicationState.Opened)
            {
                try
                {
                    OpenSession(userInfo.Name, userInfo.Password);
                }
                catch (Exception)
                {
                    throw new Exception("No s'ha pogut establir comunicació amb el servidor");
                    //Debug
                    //throw new Exception("No s'ha pogut establir comunicació amb el servidor:\n" + ex.Message);
                }
            }
        }

        public void UpdateUser(string name, string password)
        {
            CheckConnectionStatus();
            try
            {               
                //No enviem les dades amb userInfo perque potser l'actualització no es possible.
                scrabbleProxy.UpdateUser(new UserDTO() { Name = name, Password = password });

                //Si l'actualització no ha donat cap error actualitzem la info local
                userInfo.Name = name;
                userInfo.Password = password;
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void DeleteUser(bool deleteStats)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.DeleteUser(deleteStats);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<BoardDTO> GetUserBoards()
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetUserBoards();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<BoardDTO> GetBoards()
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetBoards();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<List<BoxDTO>> GetBoardBoxes(string boardId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetBoardBoxes(boardId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void SaveBoard(BoardDTO board, List<List<BoxDTO>> boxes)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.SaveBoard(board, boxes);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void DeleteBoard(BoardDTO board)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.DeleteBoard(board);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<TileSetDTO> GetUserTileSets()
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetUserTileSets();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<TileSetDTO> GetTileSets()
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetTileSets();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<TileDTO> GetTileSetTiles(string tileSetId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetTileSetTiles(tileSetId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void SaveTileSet(TileSetDTO tileSet, List<TileDTO> tiles)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.SaveTileSet(tileSet, tiles);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void DeleteTileSet(TileSetDTO tileSet)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.DeleteTileSet(tileSet);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<DictionaryInfoDTO> GetDictionaries()
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetDictionaryInfos();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void SaveDictionary(DictionaryInfoDTO dictionary)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.SaveDictionaryInfo(dictionary);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void DeleteDictionary(DictionaryInfoDTO dictionary)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.DeleteDictionaryInfo(dictionary);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public GameInfoDTO NewOnePlayerGame(GameInfoDTO gameInfo)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.NewOnePlayerGame(gameInfo);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public GameInfoDTO NewTwoPlayersGame(GameInfoDTO gameInfo)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.NewTwoPlayersGame(gameInfo);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void DeleteGame(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.DeleteGame(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public GameInfoDTO AddMeAsSecondPlayer(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.AddMeAsSecondPlayer(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<GameInfoDTO> GetAvailableGames()
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetAvailableGames();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<GameInfoDTO> GetCurrentGames()
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetCurrentUserGames();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public GameInfoDTO GetGameInfo(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetGameInfo(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<List<BoxDTO>> GetGameBoardStatus(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetGameBoardStatus(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<TileDTO> GetRack(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetGameRack(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void StartGame(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.StartGame(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public bool AttachToGame(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.AttachToGame(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<TileDTO> SwapTiles(string gameId, List<TileDTO> tiles)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.SwapTiles(gameId, tiles);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void PassTurn(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.PassTurn(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void Resign(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.Resign(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public MoveInfoDTO CheckMove(MoveDTO move)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.CheckMove(move);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public MoveInfoDTO MakeMove(MoveDTO move)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.MakeMove(move);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<string> GetValidLetters(string gameId)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetValidLetters(gameId);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<StatsHighGameDTO> GetHighestGameScores(bool onlyUserStats, bool excludeCPU)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetHighestGameScores(onlyUserStats, excludeCPU);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public List<StatsHighMoveDTO> GetHighestMoveScores(bool onlyUserStats, bool excludeCPU)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetHighestMoveScores(onlyUserStats, excludeCPU);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }


        public List<StatsGamePercentages> GetHighestWinGamesPercent(bool onlyUserStats, bool excludeCPU)
        {
            CheckConnectionStatus();
            try
            {
                return scrabbleProxy.GetHighestWinGamesPercent(onlyUserStats, excludeCPU);
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }

        public void DeleteUserStats()
        {
            CheckConnectionStatus();
            try
            {
                scrabbleProxy.DeleteUserStats();
            }
            catch (MessageSecurityException ex)
            {
                scrabbleProxy.Abort();
                if (ex.InnerException != null)
                {
                    throw new Exception(ex.InnerException.Message);
                }
                else
                {
                    throw;
                }
            }
        }
    }
}
