﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using System.ServiceModel;
using ScrabbleClient.Domain;
using ServerLibrary.Domain.Errors;
using ServerLibrary.WCF.DTO;
using ServerLibrary.Domain.Boards;
using ServerLibrary.Domain.Games;
using ServerLibrary.Domain.TileSets;

namespace ScrabbleClient.Interface
{
    /// <summary>
    /// Lógica de interacción para Game.xaml
    /// </summary>
    public partial class PlayGame : Window, IGameObserver
    {
        private bool newGame = false;
        private bool isMyTurn = false;
        private GameInfoDTO gameInfo;
        private List<List<BoxDTO>> boxes;
        private List<TileDTO> rack;

        public PlayGame()
        {
            InitializeComponent();
        }

        public void SetGame(GameInfoDTO gameInfo, bool newGame)
        {
            this.gameInfo = gameInfo;
            this.newGame = newGame;
        }

        private void btnBack_Click(object sender, RoutedEventArgs e)
        {
            this.Close();
        }

        private void frmPlayGame_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            try
            {
                UnregisterAsGameObserver();
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void frmPlayGame_Loaded(object sender, RoutedEventArgs e)
        {
            try
            {
                //Actualitzem l'estat bàsic de la partida (jugadors, punts)
                gameInfo = ClientManager.Instance.GetGameInfo(gameInfo.Id);

                //Carreguem el tauler amb les fitxes actuals
                boxes = ClientManager.Instance.GetGameBoardStatus(gameInfo.Id);

                //Carreguem el faristol
                rack = ClientManager.Instance.GetRack(gameInfo.Id);

                LoadFullGame();

                //Registrem la finestra com a observadora dels events que el server informi i que distribuïrà el clientManager
                RegisterAsGameObserver();

                //Si es una nova partida notifiquem que pot començar
                if (newGame)
                {
                    isMyTurn = false;
                    SetTurnOptions();
                    ClientManager.Instance.StartGame(gameInfo.Id);
                }
                //Si no, vinculem l'usuari a la partida i preguntem pel torn
                else
                {
                    isMyTurn = ClientManager.Instance.AttachToGame(gameInfo.Id);
                    SetTurnOptions();
                }
            }
            catch (FaultException ex)
            {
                if (ex.Code.Name == ScrabbleError.GAME_NOT_FOUND)
                {
                    MessageBox.Show(this, "No s'ha pogut trobar la partida.\nÉs possible que el rival l'hagi eliminat.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
                else
                {
                    MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void LoadFullGame()
        {
            LoadBoard();

            LoadRack();

            LoadGameStatus();
        }

        private void LoadGameStatus()
        {
            lblP1Name.Content = gameInfo.Player1Name;
            lblP1Score.Content = gameInfo.Player1Score;
            lblP2Name.Content = gameInfo.Player2Name;
            lblP2Score.Content = gameInfo.Player2Score;

            switch (gameInfo.RemainingTiles)
            {
                case 0:
                    lblTileNumber.Content = "No queda cap més fitxa al sac";
                    break;

                case 1:
                    lblTileNumber.Content = "Queda 1 fitxa al sac";
                    break;

                default:
                    lblTileNumber.Content = String.Format("Queden {0} fitxes al sac", gameInfo.RemainingTiles);
                    break;
            }

            if (gameInfo.Finished)
            {
                isMyTurn = false;
                DisableAllControls();
                lblResult.Content = "Partida finalitzada.\n";

                switch (gameInfo.WinnerPlayer)
                {
                    case 0:
                        lblResult.Content += "Heu empatat.";
                        break;

                    case 1:
                        if (ClientManager.Instance.User.Id == gameInfo.Player1)
                        {
                            lblResult.Content += "Has guanyat!";
                        }
                        else
                        {
                            lblResult.Content += "Has perdut!";
                        }
                        break;

                    case 2:
                        if (ClientManager.Instance.User.Id == gameInfo.Player2)
                        {
                            lblResult.Content += "Has guanyat!";
                        }
                        else
                        {
                            lblResult.Content += "Has perdut!";
                        }
                        break;
                }

                lblResult.Visibility = System.Windows.Visibility.Visible;
            }
        }

        private void LoadBoard()
        {
            for (int row = 0; row < Board.BOARD_LENGTH; row++)
            {
                for (int col = 0; col < Board.BOARD_LENGTH; col++)
                {
                    //Accedim a les caselles del grid per fila i columna mitjançant una consulta LINQ i l'associem a l'objecte intern
                    //I l'imatge que correspongui
                    Label box = grdBoard.Children.Cast<Label>().First(img => Grid.GetRow(img) == row && Grid.GetColumn(img) == col);

                    //Si la casella no te fitxa pintem la casella, sinó pintem la fitxa (amb la col·lecció de tons grisos)
                    BoxDTO boxDTO = boxes[row][col];
                    if (boxDTO.Tile != null)
                    {
                        ((ImageBrush)box.Background).ImageSource = GetTileImageSourceYellow(boxDTO.Tile.Value);
                        box.Content = boxDTO.Tile.Letter;
                        //Si el tauell ve amb fitxes no es pot fer ni drag ni drop desde/cap a elles
                        box.MouseLeftButtonDown -= lblBox_MouseLeftButtonDown;
                        box.Drop -= lblBox_Drop;
                        box.AllowDrop = false;
                    }
                    else
                    {
                        ((ImageBrush)box.Background).ImageSource = GetBoxImageSource((Box.BoxType)boxDTO.BoxType);
                        box.Content = "";
                        //Si es una casella lliure abilitarem el drag & drop
                        box.AllowDrop = true;
                        box.MouseLeftButtonDown += lblBox_MouseLeftButtonDown;
                        box.Drop += lblBox_Drop;
                    }

                    box.Tag = boxDTO;
                    boxDTO.IsNewMoveTile = false;
                }
            }
        }

        private void LoadRack()
        {
            for (int i = 0; i < Rack.RACK_TILES; i++)
            {
                //Accedim a les caselles del grid per columna mitjançant una consulta LINQ i l'associem a l'objecte intern
                //I l'imatge que correspongui
                Label tile = grdRack.Children.Cast<Label>().First(img => Grid.GetColumn(img) == i);

                TileDTO tileDTO = rack[i];

                if (tileDTO != null)
                {
                    tile.Content = tileDTO.Letter;
                    ((ImageBrush)tile.Background).ImageSource = GetTileImageSourceGrey(tileDTO.Value);
                }
                else
                {
                    tile.Content = "";
                    ((ImageBrush)tile.Background).ImageSource = null;
                }

                tile.Tag = rack[i];
            }
        }

        private void SetTurnOptions()
        {
            //Habilita o deshabilita les opcions segons si es el torn o no
            if (isMyTurn)
            {
                btnPlay.IsEnabled = true;
                btnSwapTiles.IsEnabled = true;
                btnPass.IsEnabled = true;

                if (ClientManager.Instance.User.Id == gameInfo.Player1)
                {
                    imgP1Turn.Source = imgStar.Source;
                    imgP2Turn.Source = imgBlank.Source;
                }
                //Assumim que si no es el jugador 1 ha de ser el 2
                else
                {
                    imgP1Turn.Source = imgBlank.Source;
                    imgP2Turn.Source = imgStar.Source;
                }
            }
            else
            {
                btnPlay.IsEnabled = false;
                btnSwapTiles.IsEnabled = false;
                btnPass.IsEnabled = false;
                //El botó per rendir-se sempre estarà habilitat

                if (ClientManager.Instance.User.Id == gameInfo.Player2)
                {
                    imgP1Turn.Source = imgStar.Source;
                    imgP2Turn.Source = imgBlank.Source;
                }
                //Assumim que si no es el jugador 2 ha de ser el 1
                else
                {
                    imgP1Turn.Source = imgBlank.Source;
                    imgP2Turn.Source = imgStar.Source;
                }
            }
        }

        private void DisableAllControls()
        {
            //Botons
            btnPlay.IsEnabled = false;
            btnSwapTiles.IsEnabled = false;
            btnPass.IsEnabled = false;
            btnResign.IsEnabled = false;
            btnRecall.IsEnabled = false;

            //tauler
            for (int i = 0; i < Board.BOARD_LENGTH; i++)
            {
                for (int j = 0; j < Board.BOARD_LENGTH; j++)
                {
                    Label box = grdBoard.Children.Cast<Label>().First(img => Grid.GetRow(img) == i && Grid.GetColumn(img) == j);
                    box.AllowDrop = false;
                    box.MouseLeftButtonDown -= lblBox_MouseLeftButtonDown;
                    box.Drop -= lblBox_Drop;
                }
            }

            //Faristol
            for (int i = 0; i < rack.Count(); i++)
            {
                Label tile = grdRack.Children.Cast<Label>().First(img => Grid.GetColumn(img) == i);
                tile.AllowDrop = false;
                tile.MouseLeftButtonDown -= lblTile_MouseLeftButtonDown;
                tile.Drop -= lblTile_Drop;
            }
        }

        private void ShowMoveInfo(MoveInfoDTO moveInfo)
        {
            if (moveInfo != null)
            {
                string formatedInfo = "";
                switch (moveInfo.Type)
                {
                    case 0:
                        string words = string.Join<string>(", ", moveInfo.Words);
                        formatedInfo = string.Format("{0} ha jugat {1} per {2} punt/s", moveInfo.PlayerName, words, moveInfo.Score);
                        break;

                    case 1:
                        formatedInfo = string.Format("{0} ha passat el torn", moveInfo.PlayerName);
                        break;

                    case 2:
                        formatedInfo = string.Format("{0} ha canviat fitxes", moveInfo.PlayerName);
                        break;

                }

                lblMoveInfo.Content = formatedInfo;
            }
        }

        public void YourTurn(GameInfoDTO gameInfo, List<List<BoxDTO>> boardInfo, List<TileDTO> rackInfo, MoveInfoDTO moveInfo)
        {
            //Nomes fem cas a la notificació si es tracta de la partida actual
            if (this.gameInfo.Id == gameInfo.Id)
            {
                ShowMoveInfo(moveInfo);

                isMyTurn = true;
                this.gameInfo = gameInfo;
                this.boxes = boardInfo;
                this.rack = rackInfo;
                LoadFullGame();

                SetTurnOptions();
            }
        }

        public void OpponentResigned(GameInfoDTO gameInfo)
        {
            //Nomes fem cas a la notificació si es tracta de la partida actual
            if (this.gameInfo.Id == gameInfo.Id)
            {
                isMyTurn = false;
                this.gameInfo = gameInfo;
                LoadFullGame();

                SetTurnOptions();
            }
        }

        //Comprova si als costats d'una casella hi ha fitxes ja usades
        private bool CheckAnchorBox(int row, int column)
        {
            //Superior
            if (row > 0 && boxes[row - 1][column].Tile != null && !boxes[row - 1][column].IsNewMoveTile)
            {
                return true;
            }

            //Inferior
            if (row < 14 && boxes[row + 1][column].Tile != null && !boxes[row + 1][column].IsNewMoveTile)
            {
                return true;
            }

            //Esquerra
            if (column > 0 && boxes[row][column - 1].Tile != null && !boxes[row][column - 1].IsNewMoveTile)
            {
                return true;
            }

            //Dreta
            if (column < 14 && boxes[row][column + 1].Tile != null && !boxes[row][column + 1].IsNewMoveTile)
            {
                return true;
            }

            return false;
        }

        private int GetValidDirection(MoveTileDTO moveTile)
        {
            int row = moveTile.Row;
            int col = moveTile.Column;

            //Si te casella a dreta o esquerra, direcció serà horitzontal
            if ((col > 0 && boxes[row][col - 1].Tile != null) ||
                (col < Board.BOARD_LENGTH - 1 && boxes[row][col + 1].Tile != null))
            {
                return 0;
            }

            //Si te casella a dalt o abaix, direcció serà vertical
            if ((row > 0 && boxes[row - 1][col].Tile != null) ||
                (row < Board.BOARD_LENGTH - 1 && boxes[row + 1][col].Tile != null))
            {
                return 1;
            }

            return -1;
        }

        private MoveDTO GetMove()
        {
            MoveDTO result = new MoveDTO();
            List<MoveTileDTO> tiles = new List<MoveTileDTO>();

            BoxDTO box;
            bool usingInitialBox = false;
            bool usingExistingTiles = false;

            //Mentre obtenim la jugada també podem fer unes primeres comprovacions (per evitar comunicació innecessaria amb el servidor)

            int firstRow = -1;
            int firstColumn = -1;
            int direction = -1;

            //Recorrem tot el tauler
            for (int row = 0; row < Board.BOARD_LENGTH; row++)
            {
                for (int col = 0; col < Board.BOARD_LENGTH; col++)
                {
                    //Obtenim la casella
                    box = boxes[row][col];

                    //Si la casella es la inicial...
                    if (box.BoxType == (int)Box.BoxType.Initial)
                    {
                        //...ha de tenir una fitxa assignada (be sigui perque ja venia d'altre jugada o perque li ha posat el jugador)
                        if (box.Tile == null)
                        {
                            throw new Exception("Jugada invàlida. S'ha d'usar la casella inicial.");
                        }
                        //Si l'estem fent servir en una nova jugada, marquem el flag corresponent
                        else if (box.IsNewMoveTile)
                        {
                            usingInitialBox = true;
                        }
                    }

                    //Si la casella te una fitxa posada pel jugador...
                    if (box.Tile != null && box.IsNewMoveTile)
                    {
                        //si es la primera que trobem, desem la fitxa i la posició on l'hem trobat
                        if (firstRow == -1) //firstRow i firstColumn canviaran a la vegada, no cal comprovar les dues
                        {
                            firstRow = row;
                            firstColumn = col;

                            usingExistingTiles |= CheckAnchorBox(row, col);
                            tiles.Add(new MoveTileDTO() { Row = row, Column = col, Tile = box.Tile, NewTile = true });
                        }
                        else
                        {
                            //si no es la primera i encara no tenim direcció definida, desem la direcció
                            if (direction == -1)
                            {
                                //Horitzontal, mateixa fila
                                if (firstRow == row)
                                {
                                    direction = 0;
                                }
                                //Vertical, mateixa columna
                                else if (firstColumn == col)
                                {
                                    direction = 1;
                                }
                                //Ni mateixa fila ni columna, jugada invalida
                                else
                                {
                                    throw new Exception("Jugada invàlida. Les fitxes han d'estar totes a la mateixa fila o columna");
                                }

                                usingExistingTiles |= CheckAnchorBox(row, col);
                                tiles.Add(new MoveTileDTO() { Row = row, Column = col, Tile = box.Tile, NewTile = true });

                            }
                            //si ja tenim primera fitxa i direccio definida comprovem que s'ajusti a la linia o columna
                            else
                            {
                                if (direction == 0 && firstRow != row)
                                {
                                    throw new Exception("Jugada invàlida. Les fitxes han d'estar totes a la mateixa fila o columna");
                                }

                                if (direction == 1 && firstColumn != col)
                                {
                                    throw new Exception("Jugada invàlida. Les fitxes han d'estar totes a la mateixa fila o columna");
                                }

                                usingExistingTiles |= CheckAnchorBox(row, col);
                                tiles.Add(new MoveTileDTO() { Row = row, Column = col, Tile = box.Tile, NewTile = true });
                            }
                        }
                    }
                }
            }
            //Arribats aquí ja hem recorregut el tauler i obtingut les fitxes noves

            //Si no hi ha cap fitxa nova la jugada no es vàlida
            if (tiles.Count() == 0)
            {
                throw new Exception("Jugada invàlida. Has d'afegir com a mínim una fitxa");
            }

            //Si es la primera jugada, hem de tenir com a minim dos fitxes
            if (usingInitialBox)
            {
                if (tiles.Count() < 2)
                {
                    throw new Exception("Jugada invàlida. La primera jugada ha de tenir com a mínim dos fitxes");
                }
            }
            //Si no es la primera jugada les fitxes han de tocar com a mínim a una fitxa ja jugada
            else
            {
                if (!usingExistingTiles)
                {
                    throw new Exception("Jugada invàlida. Com a mínim una de les noves fitxes ha d'estar adjacent a una fitxa ja jugada");
                }
            }

            //Si al arribar aqui, no hi ha direcció definida es que nomes s'ha afegit una fitxa, definim una direcció segons les fitxes adjacents
            if (direction == -1)
            {
                direction = GetValidDirection(tiles[0]);
            }


            //Ara hem de comprovar que entre totes elles no queda cap espai buit
            //També obtindrem la casella inicial de la jugada (que no es la primera fitxa de l'usuari, sino la primera de la paraula principal)
            int pos;
            bool isBlankSpace = false;

            if (direction == 0) //Horitzontal
            {
                pos = firstColumn;

                //Comprovem espais
                while (pos < Board.BOARD_LENGTH)
                {
                    //Obtenim la casella
                    box = boxes[firstRow][pos];

                    //Si trobem un espai en blanc marquem el flag
                    if (box.Tile == null)
                    {
                        isBlankSpace = true;
                    }
                    //Si encara no hem trobat cap espai en blanc i la casella on estem no es nova l'afegim a la jugada
                    else if (!isBlankSpace && !box.IsNewMoveTile)
                    {
                        tiles.Add(new MoveTileDTO() { Row = firstRow, Column = pos, Tile = box.Tile, NewTile = false });
                    }
                    //Si despres de trobar una en blanc trobem un altre fixa de l'usuari la jugada es invàlida
                    else if (isBlankSpace && box.IsNewMoveTile)
                    {
                        throw new Exception("Jugada invàlida. No pot haver-hi espais en blanc entre les teves fitxes");
                    }

                    pos++;
                }

                //Mentre hi hagi columna a l'esquerra i
                //Aquella columna tingui fitxa associada, ens movem a l'esquerra i l'afegim a la jugada
                while (firstColumn - 1 >= 0 && boxes[firstRow][firstColumn - 1].Tile != null)
                {
                    firstColumn--;

                    //Obtenim la casella
                    box = boxes[firstRow][firstColumn];

                    tiles.Add(new MoveTileDTO() { Row = firstRow, Column = firstColumn, Tile = box.Tile, NewTile = false });
                }
            }
            else //Vertical
            {
                pos = firstRow;

                //Comprovem espais
                while (pos < Board.BOARD_LENGTH)
                {
                    //Obtenim la casella
                    box = boxes[pos][firstColumn];

                    //Si trobem un espai en blanc marquem el flag
                    if (box.Tile == null)
                    {
                        isBlankSpace = true;
                    }
                    //Si encara no hem trobat cap espai en blanc i la casella on estem no es nova l'afegim a la jugada
                    else if (!isBlankSpace && !box.IsNewMoveTile)
                    {
                        tiles.Add(new MoveTileDTO() { Row = pos, Column = firstColumn, Tile = box.Tile, NewTile = false });
                    }
                    //Si despres de trobar una en blanc trobem un altre de l'usuari la jugada es invàlida
                    else if (isBlankSpace && box.IsNewMoveTile)
                    {
                        throw new Exception("Jugada invàlida. No pot haver-hi espais en blanc entre les teves fitxes");
                    }

                    pos++;
                }

                //Mentre hi hagi fila a sobre i
                //Aquella fila tingui fitxa associada, ens movem cap amunt i l'afegim a la jugada
                while (firstRow - 1 >= 0 && boxes[firstRow - 1][firstColumn].Tile != null)
                {
                    firstRow--;

                    //Obtenim la casella
                    box = boxes[firstRow][firstColumn];

                    tiles.Add(new MoveTileDTO() { Row = firstRow, Column = firstColumn, Tile = box.Tile, NewTile = false });
                }
            }

            //Si arribem aqui tenim una jugada ben construida (tot i que les paraules que formin no siguin valides i això ho validarà el servidor)
            result.GameId = gameInfo.Id;
            result.Direction = direction;
            result.InitialRow = firstRow;
            result.InitialColumn = firstColumn;
            result.Tiles = tiles;

            return result;
        }

        private void btnPlay_Click(object sender, RoutedEventArgs e)
        {
            MoveDTO move;
            MoveInfoDTO moveInfo;
            try
            {
                //Obtenim la jugada
                move = GetMove();
            }           
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            try
            {
                //Comprovem la jugada amb el servidor i demanem confirmació
                moveInfo = ClientManager.Instance.CheckMove(move);
                string words = string.Join<string>(", ", moveInfo.Words);

                if (MessageBox.Show(this, String.Format("Vols jugar {0} per {1} punts?", words, moveInfo.Score), "Confirmació", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.Yes)
                {
                    //Realitzem la jugada i canviem el torn
                    moveInfo = ClientManager.Instance.MakeMove(move);

                    isMyTurn = false;
                    SetTurnOptions();

                    ShowMoveInfo(moveInfo);

                    //Actualitzem l'estat bàsic de la partida (jugadors, punts)
                    gameInfo = ClientManager.Instance.GetGameInfo(gameInfo.Id);

                    //Carreguem el tauler amb les fitxes actuals
                    boxes = ClientManager.Instance.GetGameBoardStatus(gameInfo.Id);

                    //Carreguem el faristol
                    rack = ClientManager.Instance.GetRack(gameInfo.Id);

                    LoadFullGame();
                }
            }
            catch (FaultException ex)
            {
                if (ex.Code.Name == ScrabbleError.GAME_NOT_FOUND)
                {
                    MessageBox.Show(this, "No s'ha pogut trobar la partida.\nÉs possible que el rival l'hagi eliminat.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
                else
                {
                    MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void btnPass_Click(object sender, RoutedEventArgs e)
        {
            //Renuncia al torn i el passa a l'altre jugador

            //Demanem confirmació
            if (MessageBox.Show(this, "Estas segur que vols passar el torn?", "Confirmació", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.Yes)
            {
                try
                {
                    ClientManager.Instance.PassTurn(gameInfo.Id);

                    isMyTurn = false;
                    SetTurnOptions();

                    gameInfo = ClientManager.Instance.GetGameInfo(gameInfo.Id);

                    LoadGameStatus();
                }
                catch (FaultException ex)
                {
                    if (ex.Code.Name == ScrabbleError.GAME_NOT_FOUND)
                    {
                        MessageBox.Show(this, "No s'ha pogut trobar la partida.\nÉs possible que el rival l'hagi eliminat.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                    else
                    {
                        MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        private void btnSwapTiles_Click(object sender, RoutedEventArgs e)
        {
            //Canvia les fitxes sel·leccionades i canvia el torn

            try
            {
                //Si al sac no li queden com a minim 7 fitxes no es pot fer el canvi
                if (gameInfo.RemainingTiles < 7)
                {
                    MessageBox.Show(this, "No queden fitxes prous fitxes al sac per canviar", "Informació");
                    return;
                }

                //Comprovem que totes les fitxes estiguin al faristol
                int internalRackTileCount = 0;
                int rackTileCount = 0;

                for (int i = 0; i < 7; i++)
                {
                    if (rack[i] != null) { internalRackTileCount++; }
                    if (((Label)grdRack.Children.Cast<Label>().First(img => Grid.GetColumn(img) == i)).Tag != null) { rackTileCount++; }
                }

                if (internalRackTileCount != rackTileCount)
                {
                    MessageBox.Show(this, "Posa totes les teves fitxes al faristol per poder fer el canvi", "Informació");
                    return;
                }

                SwapTiles st = new SwapTiles();
                st.Owner = this;
                st.GameId = gameInfo.Id;
                st.Rack = rack;
                st.ShowDialog();

                //Si hem canviat fitxes, recarreguem el rack
                if (st.TilesSwaped)
                {
                    isMyTurn = false;
                    SetTurnOptions();
                    rack = st.Rack;
                    LoadRack();

                    gameInfo = ClientManager.Instance.GetGameInfo(gameInfo.Id);

                    LoadGameStatus();
                }
            }
            catch (FaultException ex)
            {
                if (ex.Code.Name == ScrabbleError.GAME_NOT_FOUND)
                {
                    MessageBox.Show(this, "No s'ha pogut trobar la partida.\nÉs possible que el rival l'hagi eliminat.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
                else
                {
                    MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void btnResign_Click(object sender, RoutedEventArgs e)
        {
            //El jugador es rendeix

            //Demanem confirmació
            if (MessageBox.Show(this, "Estas segur que vols rendir-te?", "Confirmació", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.Yes)
            {
                try
                {
                    ClientManager.Instance.Resign(gameInfo.Id);

                    isMyTurn = false;
                    SetTurnOptions();

                    gameInfo = ClientManager.Instance.GetGameInfo(gameInfo.Id);

                    LoadGameStatus();
                }
                catch (FaultException ex)
                {
                    if (ex.Code.Name == ScrabbleError.GAME_NOT_FOUND)
                    {
                        MessageBox.Show(this, "No s'ha pogut trobar la partida.\nÉs possible que el rival l'hagi eliminat.", "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                    else
                    {
                        MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(this, ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        private void btnRecall_Click(object sender, RoutedEventArgs e)
        {
            //Tornarem les fitxes posades al tauler al faristol
            for (int row = 0; row < Board.BOARD_LENGTH; row++)
            {
                for (int col = 0; col < Board.BOARD_LENGTH; col++)
                {
                    if (boxes[row][col].Tile != null && boxes[row][col].IsNewMoveTile)
                    {
                        int pos = 0;
                        bool spaceFound = false;
                        Label tile = null;

                        while (pos < Rack.RACK_TILES && !spaceFound)
                        {
                            tile = grdRack.Children.Cast<Label>().First(img => Grid.GetColumn(img) == pos);
                            if (tile.Tag == null)
                            {
                                spaceFound = true;
                            }
                            else
                            {
                                pos++;
                            }
                        }

                        //Si hem trobat un lloc lliure (sempre hauria de ser així ja que les fitxes que estan al tauler no estan al faristol)
                        //passem la fitxa del tauler al faristol (si era un escarràs el tornem a deixar blanc)
                        if (spaceFound)
                        {
                            Label box = grdBoard.Children.Cast<Label>().First(img => Grid.GetRow(img) == row && Grid.GetColumn(img) == col);

                            BoxDTO boxDTO = (BoxDTO)box.Tag;
                            TileDTO tileDTO = boxDTO.Tile;

                            boxDTO.Tile = null;
                            boxDTO.IsNewMoveTile = false;

                            if (tileDTO.Type == (int)Tile.TileType.Blank)
                            {
                                tileDTO.Letter = "";
                            }

                            //Repintem les caselles corresponents
                            box.Content = "";
                            ((ImageBrush)box.Background).ImageSource = GetBoxImageSource((Box.BoxType)boxDTO.BoxType);

                            tile.Tag = tileDTO;
                            tile.Content = tileDTO.Letter;
                            ((ImageBrush)tile.Background).ImageSource = GetTileImageSourceGrey(tileDTO.Value);

                        }
                    }
                }
            }
        }


        private void RegisterAsGameObserver()
        {
            ClientManager.Instance.AddObserver(this);
        }

        private void UnregisterAsGameObserver()
        {
            ClientManager.Instance.RemoveObserver(this);
        }

        private ImageSource GetBoxImageSource(Box.BoxType boxType)
        {
            switch (boxType)
            {
                case Box.BoxType.Normal:
                    return imgN.Source;

                case Box.BoxType.Initial:
                    return imgI.Source;

                case Box.BoxType.DoubleLetter:
                    return imgDL.Source;

                case Box.BoxType.TripleLetter:
                    return imgTL.Source;

                case Box.BoxType.DoubleWord:
                    return imgDP.Source;

                case Box.BoxType.TripleWord:
                    return imgTP.Source;

                default:
                    return null;
            }
        }

        private ImageSource GetTileImageSourceGrey(int value)
        {
            switch (value)
            {
                case 0:
                    return img0g.Source;

                case 1:
                    return img1g.Source;

                case 2:
                    return img2g.Source;

                case 3:
                    return img3g.Source;

                case 4:
                    return img4g.Source;

                case 5:
                    return img5g.Source;

                case 8:
                    return img8g.Source;

                case 10:
                    return img10g.Source;

                default:
                    return null;
            }
        }

        private ImageSource GetTileImageSourceYellow(int value)
        {
            switch (value)
            {
                case 0:
                    return img0y.Source;

                case 1:
                    return img1y.Source;

                case 2:
                    return img2y.Source;

                case 3:
                    return img3y.Source;

                case 4:
                    return img4y.Source;

                case 5:
                    return img5y.Source;

                case 8:
                    return img8y.Source;

                case 10:
                    return img10y.Source;

                default:
                    return null;
            }
        }

        /*Gestió de Drag & drop*/
        private class DragDropData
        {
            public bool SourceIsBoard { get; set; }
            public TileDTO Tile { get; set; }
            public Label SourceLabel { get; set; }
        }

        //Mètodes del tauler
        private void lblBox_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            //Obtenim l'origen del drag
            Label lbl = (Label)e.Source;

            //Primer comprovem si podem fer drag en aquesta casella del tauler
            if (((BoxDTO)lbl.Tag).Tile != null)
            {
                //Creem l'objecte amb la fitxa que s'ha de desplaçar
                DragDropData data = new DragDropData() { SourceIsBoard = true, Tile = ((BoxDTO)lbl.Tag).Tile, SourceLabel = lbl };

                //I iniciem el procés de drag
                DragDrop.DoDragDrop(lbl, data, DragDropEffects.Move);
            }
        }

        private void lblBox_DragEnter(object sender, DragEventArgs e)
        {
            //Per si hem de fer algun check
        }

        private void lblBox_Drop(object sender, DragEventArgs e)
        {
            //Obtenim el data que estem traspassant
            DragDropData data = (DragDropData)e.Data.GetData(typeof(DragDropData));

            //Obtenim els components involucrats i les seves fitxes
            Label source = data.SourceLabel;
            TileDTO sourceTile = data.Tile;

            Label dest = (Label)sender;
            TileDTO destTile = ((BoxDTO)dest.Tag).Tile;

            //Passem la fitxa (si n'hi havia) de la casella de destí a la d'origen
            if (data.SourceIsBoard)
            {
                ((BoxDTO)source.Tag).Tile = destTile;
                //Pintem la fitxa o la casella
                if (destTile != null)
                {
                    ((ImageBrush)source.Background).ImageSource = GetTileImageSourceGrey(destTile.Value);
                    source.Content = destTile.Letter;
                    ((BoxDTO)source.Tag).IsNewMoveTile = true;
                }
                else
                {
                    ((ImageBrush)source.Background).ImageSource = GetBoxImageSource(((Box.BoxType)((BoxDTO)source.Tag).BoxType));
                    source.Content = "";
                    ((BoxDTO)source.Tag).IsNewMoveTile = false;
                }
            }
            else
            {
                source.Tag = destTile;
                if (destTile != null)
                {
                    ((ImageBrush)source.Background).ImageSource = GetTileImageSourceGrey(destTile.Value);
                    //Si la casella que estem intercanviant torna al faristol i era un escarràs, tornem a deixar la fitxa en blanc
                    if (destTile.Type == (int)Tile.TileType.Blank)
                    {
                        destTile.Letter = "";
                    }
                    source.Content = destTile.Letter;
                }
                else
                {
                    ((ImageBrush)source.Background).ImageSource = null;
                    source.Content = "";
                }
            }


            //Passem la fitxa de l'origen al destí
            ((BoxDTO)dest.Tag).Tile = sourceTile;
            if (sourceTile != null)
            {
                ((ImageBrush)dest.Background).ImageSource = GetTileImageSourceGrey(sourceTile.Value);
                //Si la fitxa que estem movent es un escarràs i l'estem passant del faristol al tauler, preguntem per quina lletra volem
                if (sourceTile.Type == (int)Tile.TileType.Blank && !data.SourceIsBoard)
                {
                    SelectLetter sl = new SelectLetter();
                    sl.Owner = this;
                    sl.GameInfo = gameInfo;
                    sl.ShowDialog();
                    sourceTile.Letter = sl.SelectedLetter;
                }

                dest.Content = sourceTile.Letter;
                ((BoxDTO)dest.Tag).IsNewMoveTile = true;
            }
            else
            {
                ((ImageBrush)dest.Background).ImageSource = null;
                dest.Content = "";
                ((BoxDTO)dest.Tag).IsNewMoveTile = false;
            }

            e.Handled = true;
        }

        //Métodes del faristol
        private void lblTile_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            //Obtenim l'origen del drag
            Label lbl = (Label)e.Source;

            //Primer comprovem si podem fer drag en aquesta casella del faristol
            if (lbl.Tag != null)
            {
                //Creem l'objecte amb la fitxa que s'ha de desplaçar
                DragDropData data = new DragDropData() { SourceIsBoard = false, Tile = (TileDTO)lbl.Tag, SourceLabel = lbl };

                //I iniciem el procés de drag
                DragDrop.DoDragDrop(lbl, data, DragDropEffects.Move);
            }

        }

        private void lblTile_DragEnter(object sender, DragEventArgs e)
        {
            //Per si hem de fer algun check
        }

        private void lblTile_Drop(object sender, DragEventArgs e)
        {
            //Obtenim el data que estem traspassant
            DragDropData data = (DragDropData)e.Data.GetData(typeof(DragDropData));

            //Obtenim els components involucrats i les seves fitxes
            Label source = data.SourceLabel;
            TileDTO sourceTile = data.Tile;

            Label dest = (Label)sender;
            TileDTO destTile = (TileDTO)dest.Tag;

            //Passem la fitxa (si n'hi havia) de la casella de destí a la d'origen
            if (data.SourceIsBoard)
            {
                ((BoxDTO)source.Tag).Tile = destTile;
                //Pintem la fitxa o la casella
                if (destTile != null)
                {
                    ((ImageBrush)source.Background).ImageSource = GetTileImageSourceGrey(destTile.Value);
                    //Si la casella que estem intercanviant va al tauler i es un escarràs, demanem la lletra que ha de tenir
                    if (destTile.Type == (int)Tile.TileType.Blank)
                    {
                        SelectLetter sl = new SelectLetter();
                        sl.Owner = this;
                        sl.GameInfo = gameInfo;
                        sl.ShowDialog();
                        destTile.Letter = sl.SelectedLetter;
                    }
                    source.Content = destTile.Letter;
                    ((BoxDTO)source.Tag).IsNewMoveTile = true;
                }
                else
                {
                    ((ImageBrush)source.Background).ImageSource = GetBoxImageSource(((Box.BoxType)((BoxDTO)source.Tag).BoxType));
                    source.Content = "";
                    ((BoxDTO)source.Tag).IsNewMoveTile = false;
                }
            }
            else
            {
                source.Tag = destTile;
                if (destTile != null)
                {
                    ((ImageBrush)source.Background).ImageSource = GetTileImageSourceGrey(destTile.Value);

                    source.Content = destTile.Letter;
                }
                else
                {
                    ((ImageBrush)source.Background).ImageSource = null;
                    source.Content = "";
                }
            }


            //Passem la fitxa de l'origen al destí
            dest.Tag = sourceTile;
            if (sourceTile != null)
            {
                ((ImageBrush)dest.Background).ImageSource = GetTileImageSourceGrey(sourceTile.Value);
                //Si la fitxa que estem movent es un escarràs i l'estem passant del tauler al faristol, tornem a deixar la fitxa en blanc
                if (sourceTile.Type == (int)Tile.TileType.Blank && data.SourceIsBoard)
                {
                    sourceTile.Letter = "";
                }
                dest.Content = sourceTile.Letter;
            }
            else
            {
                ((ImageBrush)dest.Background).ImageSource = null;
                dest.Content = "";
            }

            e.Handled = true;
        }
    }
}
