﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ServerLibrary.Domain.Games;
using ServerLibrary.Domain.Errors;
using ServerLibrary.WCF.DTO;

namespace ServerLibrary.Domain.Boards
{
    public class Board
    {
        public const int BOARD_LENGTH = 15;
        public const int START_BOXES = 1;

        //Matriu de caselles
        private Box[,] boxes = new Box[BOARD_LENGTH, BOARD_LENGTH];

        public string Id { get; set; }
        public string Name { get; set; }
        public DateTime Date { get; set; }
        public Box[,] Boxes { get { return boxes; } }

        //Comprova que una jugada estigui ben formada, calcula el que valdria i les paraules que es formen
        //(No comprova que les paraules siguin vàlides)
        public void CheckMove(Move move)
        {
            string mainWord, secondaryWord;
            int tileScore = 0;
            int wordScore = 0;
            int wordFactor = 1;
            int secondaryWordScore = 0;
            int newTiles = 0;
            Move.MoveDirection perpendicularDirection;

            move.Score = 0;
            move.Words.Clear();

            //Obtenim la informació de les fitxes ordenada
            List<MoveTile> tiles = move.Tiles.OrderBy(mt => mt.Row).ThenBy(mt => mt.Column).ToList<MoveTile>();

            mainWord = "";

            if (move.Direction == Move.MoveDirection.Right)
            {
                perpendicularDirection = Move.MoveDirection.Down;
                //L'esquerra de la primera fitxa ha d'estar en lliure o ser l'inici de tauler
                if (move.InitialColumn > 0 && boxes[move.InitialRow, move.InitialColumn - 1].Tile != null)
                {
                    ScrabbleError.GetNewFaultException("Jugada invàlida. La jugada no coincideix amb el tauler", ScrabbleError.GAME_MOVE_NOT_VALID);
                }
            }
            else
            {
                perpendicularDirection = Move.MoveDirection.Right;
                //El dalt de la primera fitxa ha d'estar en lliure o ser l'inici de tauler
                if (move.InitialRow > 0 && boxes[move.InitialRow - 1, move.InitialColumn].Tile != null)
                {
                    ScrabbleError.GetNewFaultException("Jugada invàlida. La jugada no coincideix amb el tauler", ScrabbleError.GAME_MOVE_NOT_VALID);
                }
            }

            
            MoveTile moveTile = null;

            //Primer comprovem la paraula principal
            //Recorrem totes les fitxes de la jugada
            for (int i = 0; i < tiles.Count(); i++)
            {
                moveTile = tiles[i];
                tileScore = 0;

                //Concatenem la lletra de la fitxa
                mainWord += moveTile.Letter;

                CalculateMoveTileValue(moveTile, ref tileScore, ref wordFactor);

                //Sumem el valor de la fitxa a la puntuacio de la paraula
                wordScore += tileScore;

                //Si la fitxa es nova
                if (moveTile.NewTile)
                {
                    newTiles++;
                    //hem de comprovar les possibles paraules perpendiculars
                    secondaryWord = "";
                    secondaryWordScore = 0;
                    if (GetHookedWord(moveTile, perpendicularDirection, ref secondaryWordScore, ref secondaryWord))
                    {
                        move.Words.Add(secondaryWord);
                        move.Score += secondaryWordScore;
                    }
                }
            }

            //Quan tenim tota la paraula, la multipliquem pel factor de paraula que haguem acumulat
            wordScore *= wordFactor;

            move.Score += wordScore;

            //I l'afegim a la llista de paraules
            move.Words.Add(mainWord);        

            if (move.Direction == Move.MoveDirection.Right)
            {
                //La dreta de la darrera fitxa ha d'estar en lliure o ser el final de tauler
                if (moveTile.Column < BOARD_LENGTH - 1 && boxes[moveTile.Row, moveTile.Column + 1].Tile != null)
                {
                    ScrabbleError.GetNewFaultException("Jugada invàlida. La jugada no coincideix amb el tauler", ScrabbleError.GAME_MOVE_NOT_VALID);
                }
            }
            else
            {
                //El baix de la darrera fitxa ha d'estar en lliure o ser el final de tauler
                if (moveTile.Row < BOARD_LENGTH - 1 && boxes[moveTile.Row + 1, moveTile.Column].Tile != null)
                {
                    ScrabbleError.GetNewFaultException("Jugada invàlida. La jugada no coincideix amb el tauler", ScrabbleError.GAME_MOVE_NOT_VALID);
                }
            }

            //La jugada ha de tenir com a minim una fitxa nova
            if (newTiles == 0)
            {
                ScrabbleError.GetNewFaultException("Jugada invàlida. No hi ha cap fitxa nova", ScrabbleError.GAME_MOVE_NOT_VALID);
            }
            //Si s'han fet servir totes les fitxes del faristol es suma un bonus de 50 punts
            else if (newTiles == 7)
            {
                move.Score += 50;
            }
        }

        //Donada una posició calcula el que val la fitxa i actualitza el multiplicador de paraula
        private void CalculateMoveTileValue(MoveTile moveTile, ref int tileScore, ref int wordFactor)
        {
            tileScore = moveTile.Tile.Value;

            //Si es una fitxa nova, es poden aplicar modificadors de puntuació
            if (moveTile.NewTile)
            {
                switch (boxes[moveTile.Row, moveTile.Column].Type)
                {
                    case Box.BoxType.DoubleLetter:
                        tileScore *= 2;
                        break;

                    case Box.BoxType.TripleLetter:
                        tileScore *= 3;
                        break;

                    case Box.BoxType.Initial: //La casella inicial es tracta com una doble paraula 
                    case Box.BoxType.DoubleWord:
                        wordFactor *= 2;
                        break;

                    case Box.BoxType.TripleWord:
                        wordFactor *= 3;
                        break;
                }
            }
        }

        //Donada una posició i una direcció busca la possible paraula i calcula el seu valor
        private bool GetHookedWord(MoveTile moveTile, Move.MoveDirection direction, ref int score, ref string word)
        {
            int row, col;
            int tileScore = 0;
            int wordFactor = 1;
            int totalTiles = 0;
            

            if (direction == Move.MoveDirection.Right) //Horitzontal
            {
                //Busquem a esquerra i dreta fins que no hi hagi fitxes

                row = moveTile.Row;
                col = moveTile.Column -1;
                
                while (col >= 0 && boxes[row, col].Tile != null)
                {
                    score += boxes[row, col].Tile.Value;
                    word = boxes[row, col].Tile.Letter + word;
                    totalTiles++;
                    col--;
                }

                CalculateMoveTileValue(moveTile, ref tileScore, ref wordFactor);
                score += tileScore;
                word += moveTile.Letter;
                totalTiles++;

                col = moveTile.Column + 1;
                while (col < BOARD_LENGTH && boxes[row, col].Tile != null)
                {
                    score += boxes[row, col].Tile.Value;
                    word += boxes[row, col].Tile.Letter;
                    totalTiles++;
                    col++;
                }
            }
            else //Vertical
            {
                //Busquem a dalt i abaix fins que no hi hagi fitxes

                row = moveTile.Row - 1;
                col = moveTile.Column;

                while (row >= 0 && boxes[row, col].Tile != null)
                {
                    score += boxes[row, col].Tile.Value;
                    word = boxes[row, col].Tile.Letter + word;
                    totalTiles++;
                    row--;
                }

                CalculateMoveTileValue(moveTile, ref tileScore, ref wordFactor);
                score += tileScore;
                word += moveTile.Letter;
                totalTiles++;

                row = moveTile.Row + 1;
                while (row < BOARD_LENGTH && boxes[row, col].Tile != null)
                {
                    score += boxes[row, col].Tile.Value;
                    word += boxes[row, col].Tile.Letter;
                    totalTiles++;
                    row++;
                }
            }

            score *= wordFactor;

            //Si la longitut de la paraula es mes gran que 1 (no nomes hi ha la fitxa nova) tornem true
            return (totalTiles > 1);
        }

        //Afegeix les fitxes noves d'una jugada al tauler
        public void ProcessMove(Move move)
        {
            //Per cada fitxa de la jugada
            foreach (MoveTile moveTile in move.Tiles)
            {
                //Si la fitxa es nova, associem la fitxa al tauler
                if (moveTile.NewTile)
                {
                    boxes[moveTile.Row, moveTile.Column].Tile = moveTile.Tile;
                }
            }
        }

        //Comprova si unes coordenades de fila i columna cauen dins el tauler
        public void CheckValidBoxPosition(int row, int col)
        {
            if (row < 0 || row >= BOARD_LENGTH || col < 0 || col >= BOARD_LENGTH)
            {
                ScrabbleError.GetNewFaultException(String.Format("Posició invalida al tauler: Fila {0}, Columna {1}", row, col), ScrabbleError.BOARD_INVALID_BOX_POSITION);
            }
        }

        //Comprova si als costats d'una casella hi ha fitxes ja usades
        public bool CheckAnchorBox(int row, int column)
        {
            //Una casella serà tipus àncora si està lliure i te una casella amb una fitxa ja jugada adjacent
            if (boxes[row, column].Tile != null)
            {
                return false;
            }

            //Superior
            if (row > 0 && boxes[row - 1, column].Tile != null)
            {
                return true;
            }

            //Inferior
            if (row < 14 && boxes[row + 1, column].Tile != null)
            {
                return true;
            }

            //Esquerra
            if (column > 0 && boxes[row, column - 1].Tile != null)
            {
                return true;
            }

            //Dreta
            if (column < 14 && boxes[row, column + 1].Tile != null)
            {
                return true;
            }

            return false;
        }

   
        public List<List<BoxDTO>> ToDTO()
        {
            List<List<BoxDTO>> result = new List<List<BoxDTO>>();

            for (int row = 0; row < BOARD_LENGTH; row++)
            {
                List<BoxDTO> rowList = new List<BoxDTO>();

                for (int col = 0; col < BOARD_LENGTH; col++)
                {
                    rowList.Add(boxes[row, col].ToDTO());
                }

                result.Add(rowList);
            }

            return result;
        }
    }
}
