﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using System.Threading.Tasks;
using ServerLibrary.WCF.DTO;
using ServerLibrary.Domain.Boards;
using ServerLibrary.Domain.TileSets;
using ServerLibrary.Utils;
using ServerLibrary.Domain.Dictionaries;
using System.Threading;
using System.Diagnostics;
using ServerLibrary.Domain.Errors;

namespace ServerLibrary.Domain.Games
{
    /* Classe que representarà un jugador controlat per la màquina.
     * La IA de generació de moviments possibles estarà basada en el paper que es pot trobar a:
     * http://www.cs.cmu.edu/afs/cs/academic/class/15451-s06/www/lectures/scrabble.pdf
     */
    public class CPUPlayer : BasePlayer
    {
        //Classe auxiliar per desar les posicions des de les que començar a buscar jugades
        private class BoxPosition
        {
            public int Row;
            public int Col;
        }

        public const string CPUPLAYERNAME = "CPU";
        public const string CPUPLAYERID = "989dbad4-cb12-468c-a298-4411694819e9";

        //Diccionari en forma de trie que es fara servir per comprovar la validesa de les jugades que es vagin construint
        //Per la implementació d'aquesta IA es necessari un diccionari en forma d'arbre, no val una taula de hash
        private Trie trieDictionary = null;

        //Millor jugada que s'hagi trobat fins el moment
        private Move nextMove;

        //Variable per comptar quantes jugades s'han avaluat a cada torn (merament informatiu a la consola)
        private int processedMoves;

        //Flag que indicarà si es pot començar a generar una jugada o no
        private bool doMove = false;

        //Matrius de flags per saber quines caselles hem de tornar a comprovar despres d'una jugada de l'oponent
        private bool[,] recheckMatrixAcross = new bool[Board.BOARD_LENGTH, Board.BOARD_LENGTH];
        private bool[,] recheckMatrixDown = new bool[Board.BOARD_LENGTH, Board.BOARD_LENGTH];

        //Matrius de diccionaris per desar les lletres vàlides a cada posició
        private Dictionary<string, bool>[,] acrossChecks = new Dictionary<string, bool>[Board.BOARD_LENGTH, Board.BOARD_LENGTH];
        private Dictionary<string, bool>[,] downChecks = new Dictionary<string, bool>[Board.BOARD_LENGTH, Board.BOARD_LENGTH];

        //Thread que genererà la jugada
        Thread cpuThread = null;

        public override string Name { get { return CPUPLAYERNAME; } }
        public override string Id { get { return CPUPLAYERID; } }

        public override Game Game 
        { 
            set 
            { 
                game = value;
                trieDictionary = ((TrieDictionary)game.Dictionary).InternalStruct;
            } 
        }

        public CPUPlayer()
        {
            //Creem i arranquem el thread que s'encarregarà de fer les jugades
            cpuThread = new Thread(new ThreadStart(ThreadLoopControl));
            cpuThread.Start();

            //Inicialitzem les matrius
            for (int row = 0; row < Board.BOARD_LENGTH; row++)
            {
                for (int col = 0; col < Board.BOARD_LENGTH; col++)
                {
                    recheckMatrixAcross[row, col] = true;
                    recheckMatrixDown[row, col] = true;
                    acrossChecks[row, col] = new Dictionary<string, bool>();
                    downChecks[row, col] = new Dictionary<string, bool>();
                }
            }
        }

        public override void DetachUser()
        {
            cpuThread.Abort();
        }

        public override void YourTurn()
        {           
            if (game.Finished)
            {
                cpuThread.Abort();
            }
            else
            {
                //Resetejem la informació de moviments
                nextMove = null;
                lock (this) { doMove = true; }
            }
        }

        public override void OpponentResigned()
        {
            cpuThread.Abort();
        }

        private void ThreadLoopControl()
        {
            bool canWork;
            while (true)
            {
                lock (this) { canWork = doMove; }

                //Mentre no toqui fer la jugada, ens mantenim en el bucle i anem comprovant
                while (!canWork)
                {
                    Thread.Sleep(1000);

                    lock (this) { canWork = doMove; }
                }

                //Deixem un temps perque es finalitzin els processos principals de la partida
                //(acabar de retornar resultats, etc)
                Thread.Sleep(2000);

                try
                {
                    //Tornem a setejar el flag a fals
                    lock (this) { doMove = false; }
                    processedMoves = 0;
                    //Comença la magia!!
                    MakeMove();

                    //Si despres de la jugada s'ha acabat la partida sortim del thread
                    if (game.Finished)
                    {
                        return;
                    }
                }
                catch (FaultException ex)
                {
                    if (ex.Code.Name == ScrabbleError.GAME_TURN_INVALID && game.Finished)
                    {
                        return;
                    }
                    else
                    {
                        Console.WriteLine("Error processant una jugada: " + ex.Message);
                    }
                }
                catch (ThreadAbortException)
                {
                    Console.WriteLine("IA interrompuda");
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error processant una jugada: " + ex.Message);
                }
            }
        }

        //Controlarà el procés de generar una jugada vàlida
        private void MakeMove()
        {
            Stopwatch sw = Stopwatch.StartNew();
            TimeSpan elapsedTime;

            //En primer lloc recalculem les lletres vàlides de les caselles necessaries
            PrecomputeBoxes();

            //En segon lloc, obtenim totes les caselles tipus àncora
            List<BoxPosition> anchors = GetAnchorBoxes();

            nextMove = null;

            //Per cada casella àncora
            foreach (BoxPosition box in anchors)
            {
                elapsedTime = sw.Elapsed;

                //Si la comprovació porta mes de 10 segons saltem la resta 
                //(pot ser perque hi hagi molts escarrassos o per una màquina molt lenta)
                if (elapsedTime.Seconds < 10)
                {
                    //Comprovem la jugada possible horitzontal
                    CheckAcrossMove(box.Row, box.Col);

                    //Comprovem la jugada possible vertical
                    CheckDownMove(box.Row, box.Col);
                }
            }

            if (!game.Finished)
            {
                //Un cop arribats aqui, hem comprovat totes les possibles jugades i hem desat la millor
                //Si ha trobat alguna jugada, la realitzem
                if (nextMove != null)
                {
                    Console.WriteLine("CPU: Jugades avaluades en aquest torn: " + processedMoves);
                    Console.WriteLine("Jugada realitzada:");
                    Console.WriteLine(nextMove);
                    game.MakeMove(Id, nextMove);
                    //Recalculem les caselles que facin falta despres d'haver fet la jugada
                    PrecomputeBoxes();
                }
                //En el cas que no hagi trobat cap jugada (mes probable cap al final de la partida)
                else
                {
                    //Intentem canviar fitxes
                    if (game.TileSet.Tiles.Count() >= 7)
                    {
                        List<Tile> tilesToSwap = new List<Tile>();
                        for (int i = 0; i < Rack.RACK_TILES; i++)
                        {
                            Tile tile = Rack.GetTileWithoutRemove(i);
                            if (tile != null)
                            {
                                tilesToSwap.Add(tile);
                            }
                        }
                        game.SwapTiles(Id, tilesToSwap);
                    }
                    //Si no podem, passem el torn
                    else
                    {
                        game.PassTurn(Id);
                    }
                }
            }
            else
            {
                Console.WriteLine("Partida finalitzada abans de realitzar la jugada");
            }
            
            sw.Stop();
            elapsedTime = sw.Elapsed;
            Console.WriteLine("Torn processat en: " + elapsedTime);
        }

        //Recalcula les lletres vàlides per totes les caselles necessaries desde la jugada anterior
        private void PrecomputeBoxes()
        {
            //Primer actualitzem els flags de les caselles que haurem de tornar a comprovar
            if (game.Moves.Count() > 0)
            {
                Move lastMove = game.Moves.ElementAt(game.Moves.Count() - 1);

                //Per cada fitxa jugada marquem les caselles corresponents per recomprovar
                //Les caselles i direccions que s'hauran de recomprovar son:
                //Seguint la direcció vertical, tots els moviments verticals de les caselles de la columna fins trobar dos lliures seguides
                //Seguint la direccio horitzontal, totes els moviments horitzontals de les caselles de la fila fins trobar dos lliures seguides
                foreach (MoveTile moveTile in lastMove.Tiles)
                {
                    if (moveTile.NewTile)
                    {
                        int row = moveTile.Row;
                        int col = moveTile.Column;
                        int pos;
                        int consecutiveFreeBox;

                        //De la casella cap amunt
                        pos = row - 1;
                        consecutiveFreeBox = 0;
                        while (pos >= 0 && consecutiveFreeBox < 1)
                        {
                            if (game.Board.Boxes[pos, col].Tile == null)
                            {
                                recheckMatrixDown[pos, col] = true;
                                consecutiveFreeBox++;
                            }
                            else
                            {
                                consecutiveFreeBox = 0;
                            }
                            pos--;
                        }

                        //De la casella cap avall
                        pos = row + 1;
                        consecutiveFreeBox = 0;
                        while (pos < Board.BOARD_LENGTH && consecutiveFreeBox < 1)
                        {
                            if (game.Board.Boxes[pos, col].Tile == null)
                            {
                                recheckMatrixDown[pos, col] = true;
                                consecutiveFreeBox++;
                            }
                            else
                            {
                                consecutiveFreeBox = 0;
                            }
                            pos++;
                        }

                        //De la casella cap l'esquerra
                        pos = col - 1;
                        consecutiveFreeBox = 0;
                        while (pos >= 0 && consecutiveFreeBox < 1)
                        {
                            if (game.Board.Boxes[row, pos].Tile == null)
                            {
                                recheckMatrixAcross[row, pos] = true;
                                consecutiveFreeBox++;
                            }
                            else
                            {
                                consecutiveFreeBox = 0;
                            }
                            pos--;
                        }

                        //De la casella cap a la dreta
                        pos = col + 1;
                        consecutiveFreeBox = 0;
                        while (pos < Board.BOARD_LENGTH && consecutiveFreeBox < 1)
                        {
                            if (game.Board.Boxes[row, pos].Tile == null)
                            {
                                recheckMatrixAcross[row, pos] = true;
                                consecutiveFreeBox++;
                            }
                            else
                            {
                                consecutiveFreeBox = 0;
                            }
                            pos++;
                        }
                    }
                }
            }

            //Un cop sabem les caselles a recomprobar, recorrem el tauler
            for (int row = 0; row < Board.BOARD_LENGTH; row++)
            {
                for (int col = 0; col < Board.BOARD_LENGTH; col++)
                {
                    //I recalculem totes les caselles lliures que tenen el flag a true
                    if (recheckMatrixAcross[row, col] && game.Board.Boxes[row, col].Tile == null)
                    {
                        ComputeValidBoxLetters(row, col, Move.MoveDirection.Right);
                    }
                    recheckMatrixAcross[row, col] = false;

                    if (recheckMatrixDown[row, col] && game.Board.Boxes[row, col].Tile == null)
                    {
                        ComputeValidBoxLetters(row, col, Move.MoveDirection.Down);
                    }
                    recheckMatrixDown[row, col] = false;
                }
            }
        }

        //Calcula i desa les lletres vàlides per formar paraules en una direccio o altre d'una casella determinada
        private void ComputeValidBoxLetters(int row, int col, Move.MoveDirection direction)
        {
            string prefix = "";
            string postfix = "";
            bool addAll = false;

            int pos;
         
            //Recomprovem la direccio que correspongui
            if (direction == Move.MoveDirection.Right)
            {
                //Netejem la llista anterior per aquella casella
                acrossChecks[row, col].Clear();

                //Busquem els prefixes i sufixos de la casella

                //Prefix
                pos = col - 1;
                while (pos >= 0 && game.Board.Boxes[row, pos].Tile != null)
                {
                    prefix = game.Board.Boxes[row, pos].Tile.Letter + prefix;
                    pos--;
                }

                //Sufix
                pos = col + 1;
                while (pos < Board.BOARD_LENGTH && game.Board.Boxes[row, pos].Tile != null)
                {
                    postfix += game.Board.Boxes[row, pos].Tile.Letter;
                    pos++;
                }

                //Si no hi ha fitxes als costats, qualsevol fitxa serà vàlida en aquesta direcció
                if (prefix.Length == 0 && postfix.Length == 0)
                {
                    addAll = true;
                }

                //Comprovem i desem totes les lletres que facin paraules valides
                foreach (string letter in game.TileSet.ValidLetters)
                {
                    if (addAll || game.Dictionary.ContainsWord(prefix + letter + postfix))
                    {
                        acrossChecks[row, col].Add(letter, true);
                    }
                }
            }
            else
            {
                //Netejem la llista anterior per aquella casella
                downChecks[row, col].Clear();

                //Busquem els prefixes i sufixos de la casella

                //Prefix
                pos = row - 1;
                while (pos >= 0 && game.Board.Boxes[pos, col].Tile != null)
                {
                    prefix = game.Board.Boxes[pos, col].Tile.Letter + prefix;
                    pos--;
                }

                //Sufix
                pos = row + 1;
                while (pos < Board.BOARD_LENGTH && game.Board.Boxes[pos, col].Tile != null)
                {
                    postfix += game.Board.Boxes[pos, col].Tile.Letter;
                    pos++;
                }

                //Si no hi ha fitxes a dalt i baix, qualsevol fitxa serà vàlida en aquesta direcció
                if (prefix.Length == 0 && postfix.Length == 0)
                {
                    addAll = true;
                }

                //Comprovem i desem totes les lletres que facin paraules valides
                foreach (string letter in game.TileSet.ValidLetters)
                {
                    if (addAll || game.Dictionary.ContainsWord(prefix + letter + postfix))
                    {
                        downChecks[row, col].Add(letter, true);
                    }
                }
            }           
        }

        //Comprova si una lletra està dins del conjunt de lletres vàlides precalculat per una casella i direcció determinada
        private bool CheckValidBoxLetter(int row, int column, Move.MoveDirection direction, string letter)
        {
            switch (direction)
            {
                case Move.MoveDirection.Right:
                    return acrossChecks[row, column].ContainsKey(letter);

                case Move.MoveDirection.Down:
                    return downChecks[row, column].ContainsKey(letter);

                default:
                    return false;
            }
        }

        //Mètode que retornarà la llista de possibles caselles on es pot començar a buscar una jugada vàlida
        private List<BoxPosition> GetAnchorBoxes()
        {
            //Una casella ancorada es una casella lliure adjacent a una casella que ja contingui una fitxa jugada
            List<BoxPosition> result = new List<BoxPosition>();

            for (int row = 0; row < Board.BOARD_LENGTH; row++)
            {
                for (int col = 0; col < Board.BOARD_LENGTH; col++)
                {
                    if (game.Board.CheckAnchorBox(row, col))
                    {
                        result.Add(new BoxPosition() { Row = row, Col = col });
                    }
                }
            }

            //Si no hem trobat cap possible casella, es pot tractar de que no hi ha cap jugada, en aquest cas tornem la casella inicial
            if (result.Count() == 0 && game.Moves.Count() == 0)
            {
                for (int row = 0; row < Board.BOARD_LENGTH; row++)
                {
                    for (int col = 0; col < Board.BOARD_LENGTH; col++)
                    {
                        if (game.Board.Boxes[row, col].Type == Box.BoxType.Initial)
                        {
                            result.Add(new BoxPosition() { Row = row, Col = col });
                            return result;
                        }
                    }
                }
            }

            return result;
        }

        //Comprova els punts d'una jugada i la desa si es millor que l'anterior
        private void CheckLegalMove(Move move)
        {
            try
            {
                move.InitialRow = move.Tiles[0].Row;
                move.InitialColumn = move.Tiles[0].Column;
                game.CheckMove(Id, move);
                processedMoves++;

                //Debug
                Console.WriteLine("Jugada avaluada:");
                Console.WriteLine(move);

                if (nextMove != null)
                {
                    if (move.Score > nextMove.Score)
                    {
                        nextMove = move.Clone();
                    }
                }
                else
                {
                    nextMove = move.Clone();
                }
            }
            catch (FaultException ex)
            {
                if (ex.Code.Name == ScrabbleError.GAME_TURN_INVALID && game.Finished)
                {
                    throw ex;
                }
                else
                {
                    Console.WriteLine("Error processant una jugada: " + ex.Message);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error comprovant una jugada: " + ex.Message);

            }
        }

        //Processarà les possibles jugades que es poden fer de forma horitzontal partin d'una casella àncora
        private void CheckAcrossMove(int row, int col)
        {
            //Primer filtre, si aquella casella no accepta cap lletra vàlida en sentit vertical
            //O be no son vàlides cap de les que tinguem al faristol, directament descartem la casella
            if (downChecks[row, col].Count == 0 || !Rack.HasAnyValidTile(downChecks[row, col]))
            {
                return;
            }

            //Preparem l'objecte jugada
            Move move = new Move() { Direction = Move.MoveDirection.Right, Player = this };
            TrieNode node = trieDictionary.Root;

            //Si estem a l'extrem esquerra del tauler intentem extendre cap a la dreta 
            if (col == 0)
            {
                move.InitialColumn = col;
                ExtendRight(node, row, col, move, col);         
            }
            else
            {
                //Primer mirem si a l'esquerra hi ha casella lliure o ocupada
                //La part esquerra d'aquesta posició sera o tot caselles lliures o tot ocupades, pero no barrejades
                int pos = col - 1;
                //Si la part esquerra te fitxes, busquem la de mes a l'esquerra
                if (game.Board.Boxes[row, pos].Tile != null)
                {
                    while (pos - 1 >= 0 && game.Board.Boxes[row, pos - 1].Tile != null)
                    {
                        pos--;
                    }

                    bool validPrefix = true;

                    //Afegim les fitxes a la jugada i avancem en l'arbre fins la posició adient
                    for (int i = pos; i < col; i++)
                    {
                        MoveTile moveTile = new MoveTile() { Row = row, Column = i, NewTile = false, Letter = game.Board.Boxes[row, i].Tile.Letter, Tile = game.Board.Boxes[row, i].Tile };
                        move.Tiles.Add(moveTile);

                        string letter = game.Board.Boxes[row, i].Tile.Letter;

                        //La lletra d'una fitxa pot estar formada per mes d'un caracter
                        for (int charPos = 0; charPos < letter.Length; charPos++)
                        {
                            if (node != null)
                            {
                                node = node.ChildNode(game.Board.Boxes[row, i].Tile.Letter[charPos]);
                            }
                            else
                            {
                                //No s'ha arribat a un node valid de l'arbre
                                validPrefix = false;
                            }
                        }                       
                    }

                    //Intentem extendre cap a la dreta
                    if (validPrefix)
                    {
                        ExtendRight(node, row, col, move, col);
                    }
                }
                //Si te una casella lliure, mirem quants espais podem usar per crear combinacions
                //Seràn totes les caselles lliures que no siguin tipus àncora
                else
                {
                    int limit = 0;

                    //Comprovem cap endarrera per veure si encara queden mes caselles lliures
                    while (pos >= 0 && !game.Board.CheckAnchorBox(row, pos))
                    {
                        limit++;
                        pos--;
                    }

                    //Un cop sapiguem quin es el tamany màxim de part esquerra podem tenir, cridem al mètode per generar-la
                    LeftPart(node, limit, row, col, move);
                }
            }
        }

        public void LeftPart(TrieNode node, int limit, int row, int col, Move move)
        {
            //Abans d'extendre a l'esquerra actualitzem les posicions de les fitxes (van variant amb cada combinació)
            int offset = 0;
            for (int i = move.Tiles.Count - 1; i >= 0; i--)
            { 
                offset++;
                move.Tiles[i].Column = col - offset;
            }

            //Intentem extendre a la dreta la part esquerra que tenim
            ExtendRight(node, row, col, move, col);

            if (limit > 0)
            {
                //Per cada fitxa del faristol
                for (int i = 0; i < Rack.RACK_TILES; i++)
                {
                    Tile tile = Rack.GetTile(i);

                    if (tile != null)
                    {
                        //Si la fitxa es normal, pot ser que tingui mes d'un caracter (L·L, NY...)
                        if (tile.Type == Tile.TileType.Normal)
                        {
                            //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                            TrieNode childrenNode = node.ChildNode(tile.Letter[0]);
                            int index = 1;

                            while (childrenNode != null && index < tile.Letter.Count())
                            {
                                childrenNode = childrenNode.ChildNode(tile.Letter[index]);
                                index++;
                            }

                            //Hem trobat un camí valid per a tots els caràcters de la lletra
                            if (childrenNode != null)
                            {
                                //Afegim la fitxa a la jugada i seguim mirant de fer la part esquerra
                                MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                move.Tiles.Add(moveTile);

                                LeftPart(childrenNode, limit - 1, row, col, move);

                                //Quan torni, treiem la fitxa jugada
                                move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                            }
                        }
                        //Si es un escarràs, provem totes les possibles lletres vàlides
                        else
                        {
                            foreach (string letter in game.TileSet.ValidLetters)
                            {
                                //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                                TrieNode childrenNode = node.ChildNode(letter[0]);
                                int index = 1;

                                while (childrenNode != null && index < letter.Count())
                                {
                                    childrenNode = childrenNode.ChildNode(letter[index]);
                                    index++;
                                }

                                //Hem trobat un camí valid per a tots els caràcters de la lletra
                                if (childrenNode != null)
                                {
                                    //Afegim la fitxa a la jugada i seguim mirant de fer la part esquerra
                                    tile.Letter = letter;
                                    MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                    move.Tiles.Add(moveTile);

                                    LeftPart(childrenNode, limit - 1, row, col, move);

                                    //Quan torni, treiem la fitxa jugada
                                    move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                                }
                            }
                        }

                        //Tornem a posar la fitxa al faristol
                        Rack.SetTile(i, tile);
                    }
                }

            }
        }

        private void ExtendRight(TrieNode node, int row, int col, Move move, int anchorCol)
        {

            //Si no hem arribat al final del tauler
            if (col < Board.BOARD_LENGTH)
            {
                //Mirem si la casella està lliure
                if (game.Board.Boxes[row, col].Tile == null)
                {
                    //Si haviem arribat a un final de paraula vàlida, comprovem la jugada
                    //Nomes a partir del primer desplaçametn a la dreta, sino podem comprovar una paraula
                    //valida nomes amb la part esquerra
                    if (col != anchorCol && node.IsValidEnding)
                    {
                        CheckLegalMove(move);
                    }

                    //Comprovem cada fitxa del faristol si pot ser usada o no
                    for (int i = 0; i < Rack.RACK_TILES; i++)
                    {
                        Tile tile = Rack.GetTile(i);

                        if (tile != null)
                        {
                            //Si la fitxa es normal, pot ser que tingui mes d'un caracter (L·L, NY...)
                            if (tile.Type == Tile.TileType.Normal)
                            {
                                if (CheckValidBoxLetter(row, col, Move.MoveDirection.Down, tile.Letter))
                                {
                                    //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                                    TrieNode childrenNode = node.ChildNode(tile.Letter[0]);
                                    int index = 1;

                                    while (childrenNode != null && index < tile.Letter.Count())
                                    {
                                        childrenNode = childrenNode.ChildNode(tile.Letter[index]);
                                        index++;
                                    }

                                    //Hem trobat un camí valid per a tots els caràcters de la lletra
                                    if (childrenNode != null)
                                    {
                                        //Afegim la fitxa a la jugada i intentem extendre cap a la dreta
                                        MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                        move.Tiles.Add(moveTile);

                                        ExtendRight(childrenNode, row, col + 1, move, anchorCol);

                                        //Quan torni, treiem la fitxa jugada
                                        move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                                    }
                                }
                            }
                            //Si es un escarràs, provem totes les possibles lletres vàlides
                            else
                            {
                                foreach (string letter in game.TileSet.ValidLetters)
                                {
                                    if (CheckValidBoxLetter(row, col, Move.MoveDirection.Down, letter))
                                    {
                                        //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                                        TrieNode childrenNode = node.ChildNode(letter[0]);
                                        int index = 1;

                                        while (childrenNode != null && index < letter.Count())
                                        {
                                            childrenNode = childrenNode.ChildNode(letter[index]);
                                            index++;
                                        }

                                        //Hem trobat un camí valid per a tots els caràcters de la lletra
                                        if (childrenNode != null)
                                        {
                                            //Afegim la fitxa a la jugada i intentem extendre cap a la dreta
                                            tile.Letter = letter;
                                            MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                            move.Tiles.Add(moveTile);

                                            ExtendRight(childrenNode, row, col + 1, move, anchorCol);

                                            //Quan torni, treiem la fitxa jugada
                                            move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                                        }
                                    }
                                }
                            }

                            //Tornem a posar la fitxa al faristol
                            Rack.SetTile(i, tile);
                        }
                    }
                }
                //Si la casella esta ocupada
                else
                {
                    Tile tile = game.Board.Boxes[row, col].Tile;

                    //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                    TrieNode childrenNode = node.ChildNode(tile.Letter[0]);
                    int index = 1;

                    while (childrenNode != null && index < tile.Letter.Count())
                    {
                        childrenNode = childrenNode.ChildNode(tile.Letter[index]);
                        index++;
                    }

                    //Hem trobat un camí valid per a tots els caràcters de la lletra
                    if (childrenNode != null)
                    {
                        //Afegim la fitxa a la jugada i intentem extendre cap a la dreta
                        MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = false, Letter = tile.Letter, Tile = tile };
                        move.Tiles.Add(moveTile);

                        ExtendRight(childrenNode, row, col + 1, move, anchorCol);

                        //Quan torni, treiem la fitxa jugada
                        move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                    }
                }
            }
            else
            {
                //Si haviem arribat a un final de paraula vàlida, comprovem la jugada
                if (node.IsValidEnding)
                {
                    CheckLegalMove(move);
                }
            }
        }

        //Processarà les possibles jugades que es poden fer de forma vertical partin d'una casella àncora
        private void CheckDownMove(int row, int col)
        {
            //Primer filtre, si aquella casella no accepta cap lletra vàlida en sentit horitzontal
            //O be no son vàlides cap de les que tinguem al faristol, directament descartem la casella
            if (acrossChecks[row, col].Count == 0 || !Rack.HasAnyValidTile(acrossChecks[row, col]))
            {
                return;
            }

            //Preparem l'objecte jugada
            Move move = new Move() { Direction = Move.MoveDirection.Down, Player = this };
            TrieNode node = trieDictionary.Root;

            //Si estem a l'extrem superior del tauler intentem extendre cap avall 
            if (row == 0)
            {
                move.InitialRow = row;
                ExtendDown(node, row, col, move, row);
            }
            else
            {
                //Primer mirem si a sobre hi ha casella lliure o ocupada
                //La part superior d'aquesta posició sera o tot caselles lliures o tot ocupades, pero no barrejades
                int pos = row - 1;
                //Si la part superior te fitxes, busquem la de mes amunt
                if (game.Board.Boxes[pos, col].Tile != null)
                {
                    while (pos - 1 >= 0 && game.Board.Boxes[pos - 1, col].Tile != null)
                    {
                        pos--;
                    }

                    bool validPrefix = true;

                    //Afegim les fitxes a la jugada i avancem en l'arbre fins la posició adient
                    for (int i = pos; i < row; i++)
                    {
                        MoveTile moveTile = new MoveTile() { Row = i, Column = col, NewTile = false, Letter = game.Board.Boxes[i, col].Tile.Letter, Tile = game.Board.Boxes[i, col].Tile };
                        move.Tiles.Add(moveTile);

                        string letter = game.Board.Boxes[i, col].Tile.Letter;

                        //La lletra d'una fitxa pot estar formada per mes d'un caracter
                        for (int charPos = 0; charPos < letter.Length; charPos++)
                        {
                            if (node != null)
                            {
                                node = node.ChildNode(game.Board.Boxes[i, col].Tile.Letter[charPos]);
                            }
                            else
                            {
                                //No s'ha arribat a un node valid de l'arbre
                                validPrefix = false;
                            }
                        }
                    }

                    //Intentem extendre cap avall
                    if (validPrefix)
                    {
                        ExtendDown(node, row, col, move, row);
                    }
                }
                //Si te una casella lliure, mirem quants espais podem usar per crear combinacions
                //Seràn totes les caselles lliures que no siguin tipus àncora
                else
                {
                    int limit = 0;

                    //Comprovem cap endarrera per veure si encara queden mes caselles lliures
                    while (pos >= 0 && !game.Board.CheckAnchorBox(pos, col))
                    {
                        limit++;
                        pos--;
                    }

                    //Un cop sapiguem quin es el tamany màxim de part superior podem tenir, cridem al mètode per generar-la
                    AbovePart(node, limit, row, col, move);
                }
            }
        }

        public void AbovePart(TrieNode node, int limit, int row, int col, Move move)
        {
            //Abans d'extendre avall actualitzem les posicions de les fitxes (van variant amb cada combinació)
            int offset = 0;
            for (int i = move.Tiles.Count - 1; i >= 0; i--)
            {
                offset++;
                move.Tiles[i].Row = row - offset;
            }

            //Intentem extendre avall la part superior que tenim
            ExtendDown(node, row, col, move, row);

            if (limit > 0)
            {
                //Per cada fitxa del faristol
                for (int i = 0; i < Rack.RACK_TILES; i++)
                {
                    Tile tile = Rack.GetTile(i);

                    if (tile != null)
                    {
                        //Si la fitxa es normal, pot ser que tingui mes d'un caracter (L·L, NY...)
                        if (tile.Type == Tile.TileType.Normal)
                        {
                            //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                            TrieNode childrenNode = node.ChildNode(tile.Letter[0]);
                            int index = 1;

                            while (childrenNode != null && index < tile.Letter.Count())
                            {
                                childrenNode = childrenNode.ChildNode(tile.Letter[index]);
                                index++;
                            }

                            //Hem trobat un camí valid per a tots els caràcters de la lletra
                            if (childrenNode != null)
                            {
                                //Afegim la fitxa a la jugada i seguim mirant de fer la part superior
                                MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                move.Tiles.Add(moveTile);

                                AbovePart(childrenNode, limit - 1, row, col, move);

                                //Quan torni, treiem la fitxa jugada
                                move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                            }
                        }
                        //Si es un escarràs, provem totes les possibles lletres vàlides
                        else
                        {
                            foreach (string letter in game.TileSet.ValidLetters)
                            {
                                //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                                TrieNode childrenNode = node.ChildNode(letter[0]);
                                int index = 1;

                                while (childrenNode != null && index < letter.Count())
                                {
                                    childrenNode = childrenNode.ChildNode(letter[index]);
                                    index++;
                                }

                                //Hem trobat un camí valid per a tots els caràcters de la lletra
                                if (childrenNode != null)
                                {
                                    //Afegim la fitxa a la jugada i seguim mirant de fer la part superior
                                    tile.Letter = letter;
                                    MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                    move.Tiles.Add(moveTile);

                                    AbovePart(childrenNode, limit - 1, row, col, move);

                                    //Quan torni, treiem la fitxa jugada
                                    move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                                }
                            }
                        }

                        //Tornem a posar la fitxa al faristol
                        Rack.SetTile(i, tile);
                    }
                }

            }
        }

        private void ExtendDown(TrieNode node, int row, int col, Move move, int anchorRow)
        {

            //Si no hem arribat al final del tauler
            if (row < Board.BOARD_LENGTH)
            {
                //Mirem si la casella està lliure
                if (game.Board.Boxes[row, col].Tile == null)
                {
                    //Si haviem arribat a un final de paraula vàlida, comprovem la jugada
                    //Nomes a partir del primer desplaçametn avall, sino podem comprovar una paraula
                    //valida nomes amb la part superior
                    if (row != anchorRow && node.IsValidEnding)
                    {
                        CheckLegalMove(move);
                    }

                    //Comprovem cada fitxa del faristol si pot ser usada o no
                    for (int i = 0; i < Rack.RACK_TILES; i++)
                    {
                        Tile tile = Rack.GetTile(i);

                        if (tile != null)
                        {
                            //Si la fitxa es normal, pot ser que tingui mes d'un caracter (L·L, NY...)
                            if (tile.Type == Tile.TileType.Normal)
                            {
                                if (CheckValidBoxLetter(row, col, Move.MoveDirection.Right, tile.Letter))
                                {
                                    //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                                    TrieNode childrenNode = node.ChildNode(tile.Letter[0]);
                                    int index = 1;

                                    while (childrenNode != null && index < tile.Letter.Count())
                                    {
                                        childrenNode = childrenNode.ChildNode(tile.Letter[index]);
                                        index++;
                                    }

                                    //Hem trobat un camí valid per a tots els caràcters de la lletra
                                    if (childrenNode != null)
                                    {
                                        //Afegim la fitxa a la jugada i intentem extendre cap avall
                                        MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                        move.Tiles.Add(moveTile);

                                        ExtendDown(childrenNode, row + 1, col, move, anchorRow);

                                        //Quan torni, treiem la fitxa jugada
                                        move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                                    }
                                }
                            }
                            //Si es un escarràs, provem totes les possibles lletres vàlides
                            else
                            {
                                foreach (string letter in game.TileSet.ValidLetters)
                                {
                                    if (CheckValidBoxLetter(row, col, Move.MoveDirection.Right, letter))
                                    {
                                        //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                                        TrieNode childrenNode = node.ChildNode(letter[0]);
                                        int index = 1;

                                        while (childrenNode != null && index < letter.Count())
                                        {
                                            childrenNode = childrenNode.ChildNode(letter[index]);
                                            index++;
                                        }

                                        //Hem trobat un camí valid per a tots els caràcters de la lletra
                                        if (childrenNode != null)
                                        {
                                            //Afegim la fitxa a la jugada i intentem extendre cap avall
                                            tile.Letter = letter;
                                            MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = true, Letter = tile.Letter, Tile = tile };
                                            move.Tiles.Add(moveTile);

                                            ExtendDown(childrenNode, row + 1, col, move, anchorRow);

                                            //Quan torni, treiem la fitxa jugada
                                            move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                                        }
                                    }
                                }
                            }

                            //Tornem a posar la fitxa al faristol
                            Rack.SetTile(i, tile);
                        }
                    }
                }
                //Si la casella esta ocupada
                else
                {
                    Tile tile = game.Board.Boxes[row, col].Tile;

                    //Per cada caracter de la lletra, mirem si hi ha el node corresponent
                    TrieNode childrenNode = node.ChildNode(tile.Letter[0]);
                    int index = 1;

                    while (childrenNode != null && index < tile.Letter.Count())
                    {
                        childrenNode = childrenNode.ChildNode(tile.Letter[index]);
                        index++;
                    }

                    //Hem trobat un camí valid per a tots els caràcters de la lletra
                    if (childrenNode != null)
                    {
                        //Afegim la fitxa a la jugada i intentem extendre cap avall
                        MoveTile moveTile = new MoveTile() { Row = row, Column = col, NewTile = false, Letter = tile.Letter, Tile = tile };
                        move.Tiles.Add(moveTile);

                        ExtendDown(childrenNode, row + 1, col, move, anchorRow);

                        //Quan torni, treiem la fitxa jugada
                        move.Tiles.RemoveAt(move.Tiles.Count() - 1);
                    }
                }
            }
            else
            {
                //Si haviem arribat a un final de paraula vàlida, comprovem la jugada
                if (node.IsValidEnding)
                {
                    CheckLegalMove(move);
                }
            }
        }
    }
}
