﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ServerLibrary.Domain.Boards;
using ServerLibrary.Domain.Dictionaries;
using ServerLibrary.Domain.TileSets;
using ServerLibrary.LINQ;
using ServerLibrary.Domain.Users;
using ServerLibrary.Domain.Errors;
using ServerLibrary.WCF.DTO;

namespace ServerLibrary.Domain.Games
{
    public class Game
    {
        private string id;
        private string name;
        private IPlayer[] players = new IPlayer[2];
        private int turn = -1;
        private bool started = false;
        private bool finished = false;
        private int turnsWithoutScore = 0;
        private Board board = null;
        private TileSet tileSet = null;
        private IScrabbleDictionary dictionary = null;
        private List<Move> moves = null;
        private MoveInfoDTO lastMoveInfo = null;

        public Game(string id, string n, IPlayer p1, IPlayer p2, Board b, TileSet t, IScrabbleDictionary d, int tu, bool s, bool f, int tws) :
            this(id, n, p1, p2, b, t, d)
        {
            turn = tu;
            started = s;
            finished = f;
            turnsWithoutScore = tws;
        }

        public Game(string id, string n, IPlayer p1, IPlayer p2, Board b, TileSet t, IScrabbleDictionary d) :
            this(id, n, p1, b, t, d)
        {
            players[1] = p2;
        }

        public Game(string id, string n, IPlayer p1, Board b, TileSet t, IScrabbleDictionary d)
        {
            this.id = id;
            name = n;
            players[0] = p1;
            board = b;
            tileSet = t;
            dictionary = d;
        }

        public string Id { get { return id; } }
        public string Name { get { return name; } }
        public IPlayer Player1 { get { return players[0]; } set { players[0] = value; } }
        public IPlayer Player2 { get { return players[1]; } set { players[1] = value; } }
        public IPlayer TurnPlayer { get { return players[turn]; } } //Torna el jugador que te el torn
        public int Turn { get { return turn; } }
        public bool Started { get { return started; } }
        public bool Finished { get { return finished; } }
        public int TurnsWithoutScore { get { return turnsWithoutScore; } }
        public Board Board { get { return board; } }
        public TileSet TileSet { get { return tileSet; } }
        public IScrabbleDictionary Dictionary { get { return dictionary; } }
        public List<Move> Moves { get { return moves; } set { RestoreMoves(value); } }
        public MoveInfoDTO LastMoveInfo { get { return lastMoveInfo; } }

        /*
         * Operacions de previes a iniciar partida
         */

        //Donada una llista de jugades, les restaura al tauler
        private void RestoreMoves(List<Move> moves)
        {
            this.moves = moves;

            //Passar les jugades al tauler
            foreach (Move move in moves)
            {
                board.ProcessMove(move);
            }
        }

        //Afegeix un segon jugador a la partida
        public void AddPlayer2(IPlayer p2)
        {
            lock (this)
            {
                //No hi pot haver ja un segon jugador
                if (Player2 != null)
                {
                    throw ScrabbleError.GetNewFaultException("Ja hi ha un segon jugador", ScrabbleError.GAME_PLAYER2_ALREADY_EXISTS);
                }

                //No pot ser el mateix que el jugador 1
                if (Player1.Id == p2.Id)
                {
                    throw ScrabbleError.GetNewFaultException("El jugador 2 no pot ser el mateix que el jugador 1", ScrabbleError.GAME_PLAYER2_CAN_NOT_BE_PLAYER1);
                }

                Player2 = p2;
                p2.Game = this;

                GamesFactory.Instance.SavePlayer2(this);
            }
        }

        //Retorna cert si l'usuari participa en aquesta partida
        public bool IsUserInvolved(User user)
        {
            if (Player2 != null)
            {
                return (Player1.Id == user.Id || Player2.Id == user.Id);
            }
            else
            {
                return (Player1.Id == user.Id);
            }
        }

        //Vincula un usuari al seu corresponent jugador
        public bool AttachUser(User user)
        {
            lock (this)
            {
                if (Player1.Id == user.Id)
                {
                    Player1.AttachUser(user);
                }
                else if (Player2 != null && Player2.Id == user.Id)
                {
                    Player2.AttachUser(user);
                }
                else
                {
                    throw ScrabbleError.GetNewFaultException("L'usuari no està involucrat en aquesta partida", ScrabbleError.GAME_USER_NOT_INVOLVED);
                }

                return TurnPlayer.Id == user.Id;
            }
        }

        //Desvincula un usuari del seu corresponent jugador
        public void DetachUser(User user)
        {
            lock (this)
            {
                if (Player1.Id == user.Id)
                {
                    Player1.DetachUser();

                    if (Player2 != null && Player2.Id == CPUPlayer.CPUPLAYERID)
                    {
                        Player2.DetachUser();
                    }
                }
                else if (Player2 != null && Player2.Id == user.Id)
                {
                    Player2.DetachUser();
                }
                else
                {
                    throw ScrabbleError.GetNewFaultException("L'usuari no està involucrat en aquesta partida", ScrabbleError.GAME_USER_NOT_INVOLVED);
                }
            }
        }

        //Desvincula tots els usuaris 
        public void DetachAllUsers(User user)
        {
            lock (this)
            {
                //Nomes un usuari vinculat a la partida pot realitzar l'acció
                if (IsUserInvolved(user))
                {
                    Player1.DetachUser();
                    if (Player2 != null)
                    {
                        Player2.DetachUser();
                    }
                }
                else
                {
                    throw ScrabbleError.GetNewFaultException("L'usuari no està involucrat en aquesta partida", ScrabbleError.GAME_USER_NOT_INVOLVED);
                }
            }
        }

        //Comprova si hi ha usuaris vinculats als jugadors
        public bool ThereAreAttachedPlayers()
        {
            if (Player2 != null)
            {
                return (Player1.IsUserAttached() || Player2.IsUserAttached());
            }
            else
            {
                return (Player1.IsUserAttached());
            }
        }

        //Prepara tot l'entorn per a començar la partida
        public void PrepareGameStart()
        {
            //No podem començar fins que no hi hagi dos jugadors (sigui CPU o Humà)
            if (Player2 == null)
            {
                throw ScrabbleError.GetNewFaultException("No es pot començar la partida sense un segon jugador", ScrabbleError.GAME_CAN_NOT_START_WITHOUT_PLAYER2);
            }

            moves = new List<Move>();

            //Barrejem les fitxes
            tileSet.Shuffle();

            //Assignem fitxes als jugadors
            Player1.Rack.TakeTiles(tileSet);
            Player2.Rack.TakeTiles(tileSet);

            lastMoveInfo = new MoveInfoDTO() { Type = -1, Score = 0, PlayerName = "" };

            //Si la partida es d'un sol jugador, sempre comença el jugador
            if (Player2 is CPUPlayer)
            {
                turn = 0;
            }
            else
            {
                //Sinó escollim el jugador que comença (parell comença J1, imparell J2)            
                Random random = new Random();
                int nRand = random.Next(0, 100);
                turn = (nRand % 2); //Donarà 0 o 1
            }
            started = true;

            //Desem el nou estat
            GamesFactory.Instance.UpdateGame(this);
        }


        //Notifica al jugador que te el torn que li toca
        public void Start(string userId)
        {
            TurnPlayer.YourTurn();
        }


        /*
         * Operacions de joc de la partida
         */

        //Comprova si qui està intentant fer alguna acció li correspon realment el torn
        private void CheckValidTurn(string playerId)
        {
            if (TurnPlayer.Id != playerId)
            {
                throw ScrabbleError.GetNewFaultException("No es el torn d'aquest jugador", ScrabbleError.GAME_TURN_INVALID);
            }
        }

        //Provoca que el jugador renunciï al torn
        public void PassTurn(string playerId)
        {
            lock (this)
            {
                //Comprovem si es el torn del jugador
                CheckValidTurn(playerId);

                turnsWithoutScore++;
                lastMoveInfo = new MoveInfoDTO() { Type = 1, PlayerName = TurnPlayer.Name };
                CheckGameFinished();
            }
        }

        //Permet a un jugador rendir-se i donar guanyador a l'altre
        public void Resign(string playerId)
        {
            lock (this)
            {
                if (Player1.Id == playerId)
                {
                    Player1.Score = 0;
                }
                else if (Player2.Id == playerId)
                {
                    Player2.Score = 0;
                }
                else
                {
                    throw ScrabbleError.GetNewFaultException("L'usuari no està involucrat en aquesta partida", ScrabbleError.GAME_USER_NOT_INVOLVED);
                }

                GameResigned(playerId);
            }
        }

        //Canvia un conjunt de fitxes per unes de noves del sac (serà cridat pel client)
        public List<TileDTO> SwapTiles(string playerId, List<TileDTO> tiles)
        {
            lock (this)
            {

                //Comprovem si es el torn del jugador
                CheckValidTurn(playerId);

                //Treiem les fitxes solicitades del faristol del jugador
                List<Tile> oldTiles = TurnPlayer.Rack.GetTiles(tiles);

                //Omplim el faristol amb les noves fitxes
                TurnPlayer.Rack.TakeTiles(tileSet);

                //Posem les fitxes velles al sac 
                tileSet.AddTiles(oldTiles);
                tileSet.Shuffle();

                //Acabem d'omplir el faristol (si fa falta) amb el sac
                //(no faria falta, perque no es pot canviar fitxes si en queden menys de 7 al sac,
                //amb el que sempre hi ha d'haver fitxes suficients com per canviar tot el faristol,
                //però si algun dia canvia aquesta regla, així ens assegurem que el jugador te el faristol 
                //el mes plé possible)
                TurnPlayer.Rack.TakeTiles(tileSet);

                List<TileDTO> result = TurnPlayer.Rack.ToDTO();

                turnsWithoutScore++;
                lastMoveInfo = new MoveInfoDTO() { Type = 2, PlayerName = TurnPlayer.Name };
                CheckGameFinished();

                return result;
            }
        }

        //Canvia un conjunt de fitxes per unes de noves del sac (serà cridat per la IA)
        public void SwapTiles(string playerId, List<Tile> tiles)
        {
            lock (this)
            {
                //Comprovem si es el torn del jugador
                CheckValidTurn(playerId);

                //Treiem les fitxes solicitades del faristol del jugador
                List<Tile> oldTiles = new List<Tile>();
                foreach (Tile tile in tiles)
                {
                    if (tile != null)
                    {
                        oldTiles.Add(TurnPlayer.Rack.GetTile(tile.Id));
                    }
                }

                //Omplim el faristol amb les noves fitxes
                TurnPlayer.Rack.TakeTiles(tileSet);

                //Posem les fitxes velles al sac 
                tileSet.AddTiles(oldTiles);
                tileSet.Shuffle();

                //Acabem d'omplir el faristol (si fa falta) amb el sac
                //(no faria falta, perque no es pot canviar fitxes si en queden menys de 7 al sac,
                //amb el que sempre hi ha d'haver fitxes suficients com per canviar tot el faristol,
                //però si algun dia canvia aquesta regla, així ens assegurem que el jugador te el faristol 
                //el mes plé possible)
                TurnPlayer.Rack.TakeTiles(tileSet);

                turnsWithoutScore++;
                lastMoveInfo = new MoveInfoDTO() { Type = 2, PlayerName = TurnPlayer.Name };
                CheckGameFinished();
            }
        }


        //Donat una jugada en format DTO, la converteix al objecte del domini i comprova si es vàlida
        //En cas afirmatiu tornarà un objecte jugada (Move) amb els punts que val, i una llista de paraules formades
        //En cas negatiu tornarà una excepció amb la llista de paraules invàlides
        private Move CheckMoveFromDTO(string playerId, MoveDTO moveDTO)
        {
            //Comprovem si es el torn del jugador
            CheckValidTurn(playerId);

            List<string> invalidWords = new List<string>();

            Move move = new Move();
            move.FromDTO(moveDTO, TurnPlayer.Rack, board);

            //Obtenim totes les paraules a comprovar i els punts que valdria la jugada en cas que fos vàlida
            board.CheckMove(move);

            //Comprovem totes les possibles paraules de la jugada i els seus punts
            foreach (string word in move.Words)
            {
                if (!dictionary.ContainsWord(word))
                {
                    invalidWords.Add(word);
                }
            }

            //Si alguna incorrecte tornem la llista de incorrectes
            if (invalidWords.Count() > 0)
            {
                throw ScrabbleError.GetNewFaultException("Les següents paraules no són vàlides:\n" + string.Join<string>(", ", invalidWords), ScrabbleError.GAME_MOVE_NOT_VALID);
            }

            return move;
        }

        //Mètode públic que comprova una jugada en format DTO (serà cridat pels clients)
        public MoveInfoDTO CheckMove(string playerId, MoveDTO moveDTO)
        {
            lock (this)
            {
                Move move = CheckMoveFromDTO(playerId, moveDTO);

                //Si tot es ok tornem la informació de la jugada
                return new MoveInfoDTO() { Type = 0, PlayerName = TurnPlayer.Name, Score = move.Score, Words = move.Words };
            }
        }

        //Mètode públic que comprova una jugada (serà cridat per la IA)
        public void CheckMove(string playerId, Move move)
        {
            lock (this)
            {
                //Comprovem si es el torn del jugador
                CheckValidTurn(playerId);

                board.CheckMove(move);
            }
        }

        //Afegeix una jugada a la partida
        private void AddMove(Move move)
        {
            move.Player = TurnPlayer;

            //Ens assegurem que els escarrassos tenen la lletra correcta assignada
            //La CPU la pot haver canviat mentre buscava jugades, la que ha de valdre es la de moveTile
            foreach (MoveTile moveTile in move.Tiles)
            {
                if (moveTile.Tile.Type == Tile.TileType.Blank)
                {
                    moveTile.Tile.Letter = moveTile.Letter;
                }
            }

            //Actualitzem la puntuació del jugador
            TurnPlayer.Score += move.Score;

            //Afegim les fitxes al tauler
            board.ProcessMove(move);

            //Treiem les fitxes del faristol
            TurnPlayer.Rack.ProcessMove(move);

            //Assignen fitxes noves al jugador
            TurnPlayer.Rack.TakeTiles(tileSet);

            //Afegim la jugada a la partida 
            moves.Add(move);

            turnsWithoutScore = 0;

            //Desem la jugada a base de dades
            GamesFactory.Instance.SaveMove(this);

            lastMoveInfo = new MoveInfoDTO() { Type = 0, PlayerName = TurnPlayer.Name, Score = move.Score, Words = move.Words };

            CheckGameFinished();
        }

        //Mètode públic que realitza una jugada en format DTO (serà cridat pels clients)
        public MoveInfoDTO MakeMove(string playerId, MoveDTO moveDTO)
        {
            lock (this)
            {
                //Comprovem i obtenim el moviment
                Move move = CheckMoveFromDTO(playerId, moveDTO);

                AddMove(move);

                return lastMoveInfo;
            }
        }

        //Mètode públic que realitza una jugada (serà cridat per la IA)
        public void MakeMove(string playerId, Move move)
        {
            lock (this)
            {
                //Comprovem si es el torn del jugador
                CheckValidTurn(playerId);

                AddMove(move);
            }
        }

        //Comprova si la partida ha acabat
        private void CheckGameFinished()
        {
            /*
             * La partida s'acaba quan:
             *  1- El jugador que ha fet la jugada ja no te mes fitxes (després de reomplir del sac)
             *  2- S'han fet 6 jugades consecutives de 0 punts (en cas de 0-0 guanya el que el seu faristol suma menys) 
             * */

            if (TurnPlayer.Rack.Count() == 0 || turnsWithoutScore == 6)
            {
                GameFinished();
            }
            else
            {
                ChangeTurn();
            }
        }

        //Gestiona una partida acabada
        private void GameFinished()
        {
            /* Si la partida s'ha acabat s'ha de:
             *  1- Reduir la puntuació de cada jugador la suma de les seves lletres no jugades del faristol
             *  2- Si un jugador ha fet servir totes les seves lletres, se li suma el total de lletres no jugades de l'altre jugador
             *  
             * 3- Informar als jugadors de la finalització
             **/
            if (Player1.Rack.Count() == 0)
            {
                Player1.Score += Player2.Rack.GetTotalPoints();
            }
            else
            {
                Player1.Score -= Player1.Rack.GetTotalPoints();
            }

            if (Player2.Rack.Count() == 0)
            {
                Player2.Score += Player1.Rack.GetTotalPoints();
            }
            else
            {
                Player2.Score -= Player2.Rack.GetTotalPoints();
            }

            finished = true;

            ChangeTurn();
        }

        //Gestiona la rendició d'un jugador
        private void GameResigned(string playerId)
        {
            finished = true;

            //Desa l'estat
            GamesFactory.Instance.UpdateGame(this);

            //Informa al rival que el jugador s'ha rendit
            if (Player1.Id == playerId)
            {
                Player2.OpponentResigned();
            }
            else if (Player2.Id == playerId)
            {
                Player1.OpponentResigned();
            }
        }

        //Gestiona el canvi de torn
        private void ChangeTurn()
        {
            //Canvia el torn
            turn = (turn + 1) % 2;

            //Desa l'estat
            GamesFactory.Instance.UpdateGame(this);

            //Informa al jugador que li toca
            TurnPlayer.YourTurn();
        }


        /*
         * Operacions de consulta de l'estat de la partida
         */

        public GameInfoDTO ToDTO()
        {
            GameInfoDTO result = new GameInfoDTO();

            result.Id = this.id;
            result.Name = this.name;
            result.BoardName = board.Name;
            result.Board = board.Id;
            result.TileSetName = tileSet.Name;
            result.TileSet = tileSet.Id;
            result.RemainingTiles = tileSet.Tiles.Count();
            result.DictionaryName = dictionary.Name;
            result.Dictionary = dictionary.Id;
            result.Turn = turn;
            result.Player1Name = Player1.Name;
            result.Player1 = Player1.Id;
            result.Player1Score = Player1.Score;
            if (Player2 != null)
            {
                result.Player2Name = Player2.Name;
                result.Player2 = Player2.Id;
                result.Player2Score = Player2.Score;
            }
            else
            {
                result.Player2Name = "";
                result.Player2 = "";
                result.Player2Score = 0;
            }
            result.Started = this.started;
            result.Finished = this.finished;
            if (finished)
            {
                if (Player1.Score == Player2.Score)
                    result.WinnerPlayer = 0;
                else if (Player1.Score > Player2.Score)
                    result.WinnerPlayer = 1;
                else
                    result.WinnerPlayer = 2;
            }

            return result;
        }

        public List<List<BoxDTO>> GetBoardStatus()
        {
            return board.ToDTO();
        }

        public List<TileDTO> GetPlayerRack(string playerId)
        {
            if (Player1.Id == playerId)
            {
                return Player1.Rack.ToDTO();
            }
            else if (Player2.Id == playerId)
            {
                return Player2.Rack.ToDTO();
            }
            else
            {
                throw ScrabbleError.GetNewFaultException("L'usuari no està involucrat en aquesta partida", ScrabbleError.GAME_USER_NOT_INVOLVED);
            }
        }
    }
}
