﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ServerLibrary.WCF.DTO;
using ServerLibrary.LINQ;
using ServerLibrary.Domain.Users;
using ServerLibrary.Domain.Errors;

namespace ServerLibrary.Domain.Games
{
    /*
     * Classe que s'encarrega de la gestió de les partides
     */
    public sealed class GameManager
    {
        /*Gestió del singleton*/
        private static readonly GameManager instance = new GameManager();

        public static GameManager Instance
        {
            get
            {
                return instance;
            }
        }

        private GameManager() { }


        /*Variables d'instancia*/
        //Diccionari que contindrà les partides en joc carregades actualment
        private Dictionary<string, Game> currentGames = new Dictionary<string, Game>();

        public GameInfoDTO NewOnePlayerGame(GameInfoDTO gameInfo, User user)
        {
            Game newGame = GamesFactory.Instance.GetNewOnePlayerGame(gameInfo, user);

            //Com que es per un sol jugador, ja podem realitzar els passos previs
            newGame.PrepareGameStart();

            lock (this)
            {
                currentGames.Add(newGame.Id, newGame);
            }

            return newGame.ToDTO();           
        }

        public GameInfoDTO NewTwoPlayersGame(GameInfoDTO gameInfo, User user)
        {
            Game newGame = GamesFactory.Instance.GetNewTwoPlayerGame(gameInfo, user);

            lock (this)
            {
                currentGames.Add(newGame.Id, newGame);
            }

            return newGame.ToDTO();
        }

        //Esborra una partida tant de memoria com de base de dades
        public void DeleteGame(string gameId, User user)
        {
            //En aquest cas el lock ha de ser durant mes instruccions, doncs ens hem d'assegurar que mentre estem 
            //Eliminant de memoria i base de dades ningú accedeix a modificar la partida
            lock (this)
            {
                Game game = GetGame(gameId);

                game.DetachAllUsers(user);
                currentGames.Remove(game.Id);

                GamesFactory.Instance.DeleteGame(game);
            }
        }

        //Esborra totes les partides en que participa un usuari (es crida quan s'està eliminant l'usuari)
        public void DeleteAllGamesFromUser(User user)
        {
            int i = 0;
            Game game;

            lock (this)
            {
                //Borrem les partides carregades en memoria
                while (i < currentGames.Count())
                {
                    game = currentGames.ElementAt(i).Value;

                    if (game.IsUserInvolved(user))
                    {
                        DeleteGame(game.Id, user);
                    }
                    else
                    {
                        i++;
                    }
                }

                //I totes les que quedessin no carregades
                GamesFactory.Instance.DeleteAllGamesFromUser(user.Id);
            }
        }

        //Afegeix el segon jugador a una partida
        public GameInfoDTO AddSecondPlayer(string gameId, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            //I finalment afegim el jugador
            game.AddPlayer2(new HumanPlayer(user, user.Id, user.Name));

            //Com que ja tenim els dos jugadors, ja podem realitzar els passos previs
            game.PrepareGameStart();

            return game.ToDTO();
        }

        public List<GameInfoDTO> GetAvailableGames(User user)
        {
            return GamesFactory.Instance.GetAvailableGames(user.Id);
        }

        public List<GameInfoDTO> GetCurrentUserGames(User user)
        {
            return GamesFactory.Instance.GetCurrentUserGames(user.Id);
        }

        private Game GetGame(string gameId)
        {
            Game game = null;

            lock (this)
            {
                //La busquem en memoria
                if (currentGames.ContainsKey(gameId))
                {
                    game = currentGames[gameId];
                }
                //Si no hi es la carreguem de la base de dades
                else
                {
                    game = GamesFactory.Instance.LoadGame(gameId);
                    currentGames.Add(game.Id, game);
                }
            }

            return game;
        }

        //Torna l'estat d'una partida
        public GameInfoDTO GetGameInfo(string gameId)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            return game.ToDTO();
        }

        //Torna l'estat del tauler
        public List<List<BoxDTO>> GetGameBoardStatus(string gameId)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            return game.GetBoardStatus();
        }

        public List<TileDTO> GetGameRack(string gameId, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            return game.GetPlayerRack(user.Id);
        }

        public void StartGame(string gameId, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            game.Start(user.Id);
        }

        public bool AttachToGame(string gameId, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            return game.AttachUser(user);
        }

        //Desvincula el jugador de totes les partides en memoria en les que estigues involucrat
        //Si alguna partida es queda sensa cap jugador, no cal que la mantinguem en memoria
        public void DetachFromAllGames(User user)
        {
            int i = 0;
            Game game;

            while (i < currentGames.Count())
            {
                game = currentGames.ElementAt(i).Value;

                if (game.IsUserInvolved(user))
                {
                    game.DetachUser(user);
                }

                if (!game.ThereAreAttachedPlayers())
                {
                    currentGames.Remove(game.Id);
                }
                else
                {
                    i++;
                }
            }
        }

        public void PassTurn(string gameId, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            game.PassTurn(user.Id);
        }

        public List<TileDTO> SwapTiles(string gameId, User user, List<TileDTO> tiles)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            return game.SwapTiles(user.Id, tiles);
        }

        public List<string> GetValidLetters(string gameId, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            return game.TileSet.ValidLetters;
        }

        public void Resign(string gameId, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(gameId);

            game.Resign(user.Id);
        }

        public MoveInfoDTO CheckMove(MoveDTO move, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(move.GameId);

            return game.CheckMove(user.Id, move);
        }

        public MoveInfoDTO MakeMove(MoveDTO move, User user)
        {
            //En primer lloc hem de localitzar la partida.
            Game game = GetGame(move.GameId);

            return game.MakeMove(user.Id, move);
        }
    }
}
