﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;
using System.Threading.Tasks;
using ServerLibrary.WCF.DTO;
using ServerLibrary.Domain.Boards;
using ServerLibrary.Domain.Users;
using ServerLibrary.Domain.Errors;
using ServerLibrary.LINQ.Context;
using ServerLibrary.LINQ.Entity;
using System.ServiceModel;

namespace ServerLibrary.LINQ
{
    /*
     * Classe que s'encarrega de la gestió i la persistencia dels taulers.
     */
    public sealed class BoardsFactory
    {
        /*Gestió del singleton*/
        private static readonly BoardsFactory instance = new BoardsFactory();

        public static BoardsFactory Instance
        {
            get
            {
                return instance;
            }
        }

        private BoardsFactory() 
        {
            try
            {
                string connStr = ConfigurationManager.ConnectionStrings["ServerLibrary.Properties.Settings.SCRABBLEConnectionString"].ConnectionString;
                dataContext = new ScrabbleClassesDataContext(connStr);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        /*Constants*/        
        private const int BOARD_NAME_MAX_LENGTH = 100;

        /*Variables d'instancia*/

        //Classe per accedir al sistema LINQ que connecta amb la base de dades
        private ScrabbleClassesDataContext dataContext;

        /*Mètode que retorna un tauler amb les seves caselles*/
        public Board GetBoard(string boardId)
        {
            try
            {
                Board result = null;

                var dbBoards = (from b in dataContext.BOARD
                                where b.id == boardId
                                select b);

                //Si existeix, creem l'objecte i omplim les caselles
                if (dbBoards.Count() > 0)
                {
                    BOARD dbBoard = dbBoards.First();
                    result = new Board() { Id = dbBoard.id, Name = dbBoard.name, Date = dbBoard.creationDate };

                    var dbBoxes = (from bx in dbBoard.BOX
                                   orderby bx.boxRow, bx.boxColumn
                                   select bx);

                    foreach (BOX dbBox in dbBoxes)
                    {
                        Box box = new Box() { Id = dbBox.id, Type = (Box.BoxType)dbBox.type };
                        result.Boxes[dbBox.boxRow, dbBox.boxColumn] = box;
                    }
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }


        /*
         * Métodes per gestionar els taulers a través dels clients (es reben i transmeten les classes ...DTO que encapsulen les dades necessaries) 
         */
        public List<BoardDTO> GetUserBoards(string userId)
        {
            try
            {
                List<BoardDTO> result = new List<BoardDTO>();

                var dbBoards = (from b in dataContext.BOARD
                                where b.owner == userId
                                orderby b.name
                                select b);

                foreach (BOARD b in dbBoards)
                {
                    BoardDTO board = new BoardDTO() { Owner = b.owner, Id = b.id, Name = b.name, Date = b.creationDate };
                    result.Add(board);
                }
                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }


        public List<BoardDTO> GetBoards(string userId)
        {
            try
            {
                List<BoardDTO> result = new List<BoardDTO>();

                var dbBoards = (from b in dataContext.BOARD
                                where b.owner == userId || b.owner == null
                                orderby b.owner, b.name
                                select b);

                foreach (BOARD b in dbBoards)
                {
                    BoardDTO board = new BoardDTO() { Owner = b.owner, Id = b.id, Name = b.name, Date = b.creationDate };
                    result.Add(board);
                }
                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }


        public List<List<BoxDTO>> GetBoardBoxes(string boardId)
        {
            try
            {
                List<List<BoxDTO>> result = new List<List<BoxDTO>>();
                List<BoxDTO> row = new List<BoxDTO>();
                int currentRow = -1;

                var dbBoxes = (from b in dataContext.BOX
                               where b.board == boardId
                               orderby b.boxRow, b.boxColumn
                               select b);

                foreach (BOX b in dbBoxes)
                {
                    if (currentRow != b.boxRow)
                    {
                        row = new List<BoxDTO>();
                        result.Add(row);

                        currentRow = b.boxRow;
                    }

                    BoxDTO box = new BoxDTO() { Id = b.id, BoxType = (int)b.type };
                    row.Add(box);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }


        public void SaveBoard(BoardDTO board, List<List<BoxDTO>> boxes, string userId)
        {
            if (board.Name.Length > BOARD_NAME_MAX_LENGTH)
            {
                throw ScrabbleError.GetNewFaultException(String.Format("No es pot desar el tauler, el nom no pot superar els {0} caracters", BOARD_NAME_MAX_LENGTH), ScrabbleError.BOARD_NAME_TOO_LONG);
            }

            bool newBoard = false;

            //Si l'id està buit es un nou tauler i assignem un nou id
            if (board.Id == "")
            {
                board.Id = Guid.NewGuid().ToString();
                newBoard = true;
            }

            //Abans de desar s'han de fer els checks pertinents
            //1º el nom no pot estar repetit (dins els tauler del propi usuari)
            int count = 0;

            try
            {
                count = (from b in dataContext.BOARD
                         where (b.name == board.Name) && (b.owner == userId) && (b.id != board.Id)
                         select b).Count();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("No es pot desar el tauler, ja hi ha un tauler amb aquest nom.", ScrabbleError.BOARD_NAME_ALREADY_EXISTS);
            }
       
            //2º el tipus de les caselles han de ser correctes
            //3º hi ha d'haver exactament 15 files (primer nivell de l'array) i 15 columnes (segon nivell de l'array)
            //4º hi ha d'haver una i nomes una casella d'inici
            int totalRows = 0;
            int totalCols = 0;
            int startBoxes = 0;

            foreach (List<BoxDTO> row in boxes)
            {
                totalRows++;
                totalCols = 0;
                foreach (BoxDTO box in row)
                {
                    totalCols++;
                    if (!Enum.IsDefined(typeof(Box.BoxType), box.BoxType))
                    {
                        throw ScrabbleError.GetNewFaultException("No es pot desar el tauler, hi ha alguna casella invàlida.", ScrabbleError.BOARD_HAS_INVALID_BOX);
                    }

                    if (((Box.BoxType)box.BoxType) == Box.BoxType.Initial)
                    {
                        startBoxes++;
                    }
                }

                if (totalCols != Board.BOARD_LENGTH)
                {
                    throw ScrabbleError.GetNewFaultException("No es pot desar el tauler, número de columnes invalid.", ScrabbleError.BOARD_HAS_INVALID_COLUMNS_LENGTH);
                }
            }

            if (totalRows != Board.BOARD_LENGTH)
            {
                throw ScrabbleError.GetNewFaultException("No es pot desar el tauler, número de files invalid.", ScrabbleError.BOARD_HAS_INVALID_ROWS_LENGTH);
            }

            if (startBoxes != Board.START_BOXES)
            {
                throw ScrabbleError.GetNewFaultException("No es pot desar el tauler, hi ha d'haver una i nomes una casella inicial.", ScrabbleError.BOARD_HAS_INVALID_START_BOXES_NUMBER);
            }

            if (newBoard)
            {
                SaveNewBoard(board, boxes, userId);
            }
            else
            {
                SaveUpdatedBoard(board, boxes, userId);
            }
        }

        private void SaveNewBoard(BoardDTO board, List<List<BoxDTO>> boxes, string userId)
        {
            try
            {
                //Inserim el tauler
                BOARD dbBoard = new BOARD();
                dbBoard.id = board.Id;
                dbBoard.name = board.Name;
                dbBoard.owner = userId;
                dbBoard.creationDate = DateTime.UtcNow;

                //Inserim les caselles
                int currentRow = 0;
                int currentCol = 0;
                foreach (List<BoxDTO> row in boxes)
                {
                    currentCol = 0;
                    foreach (BoxDTO box in row)
                    {
                        BOX dbBox = new BOX();
                        dbBox.id = Guid.NewGuid().ToString();
                        dbBox.type = box.BoxType;
                        dbBox.boxRow = currentRow;
                        dbBox.boxColumn = currentCol;

                        dbBoard.BOX.Add(dbBox);
                        currentCol++;
                    }

                    currentRow++;
                }

                dataContext.BOARD.InsertOnSubmit(dbBoard);
                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void SaveUpdatedBoard(BoardDTO board, List<List<BoxDTO>> boxes, string userId)
        {
            //No es pot modificar si està en us en alguna partida
            CheckIfInUse(board);

            try
            {
                //Actualitzem el tauler
                var dbBoard = (from b in dataContext.BOARD
                               where b.id == board.Id
                               select b).First();

                dbBoard.name = board.Name;

                //Actualitzem les caselles
                var dbBoxes = (from bx in dbBoard.BOX
                               orderby bx.boxRow, bx.boxColumn
                               select bx);

                foreach (BOX dbBox in dbBoxes)
                {
                    dbBox.type = boxes[dbBox.boxRow][dbBox.boxColumn].BoxType;
                }

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void DeleteBoard(BoardDTO board, string userId)
        {
            //No es pot borrar si s'està fent servir en alguna partida
            CheckIfInUse(board);

            try
            {
                //Obtenim el objecte tauler
                var dbBoard = (from b in dataContext.BOARD
                               where b.id == board.Id && b.owner == userId //Els GUIDs haurien de ser unics, pero forçem a comprobar també el usuari
                               select b).First();

                dataContext.BOARD.DeleteOnSubmit(dbBoard);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void DeleteAllUserBoards(string userId)
        {
            try
            {
                //Obtenim els objectes taulers
                var dbBoards = (from b in dataContext.BOARD
                               where b.owner == userId 
                               select b);

                dataContext.BOARD.DeleteAllOnSubmit(dbBoards);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void CheckIfInUse(BoardDTO board)
        {
            int count;

            try
            {
                count = (from g in dataContext.GAME
                             where (g.BOARD1.id == board.Id) && (g.finished == false)
                             select g).Count();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("Operació no vàlida, el tauler està en ús", ScrabbleError.BOARD_IS_IN_USE);
            }
        }
    }
}
