﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Configuration;
using System.Threading.Tasks;
using ServerLibrary.LINQ.Context;
using ServerLibrary.LINQ.Entity;
using System.ServiceModel;
using ServerLibrary.WCF.DTO;
using ServerLibrary.Domain.Dictionaries;
using ServerLibrary.Domain.Errors;

namespace ServerLibrary.LINQ
{
    /*
     * Classe que s'encarrega de la gestió i la persistencia dels diccionaris.
     */
    public sealed class DictionariesFactory
    {
        /*Gestió del singleton*/
        private static readonly DictionariesFactory instance = new DictionariesFactory();

        public static DictionariesFactory Instance
        {
            get
            {
                return instance;
            }
        }

        private DictionariesFactory() 
        {
            try
            {
                string connStr = ConfigurationManager.ConnectionStrings["ServerLibrary.Properties.Settings.SCRABBLEConnectionString"].ConnectionString;
                dataContext = new ScrabbleClassesDataContext(connStr);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        /*Constants*/
        private enum DictionaryType { BasicDictionary = 0, TrieDictionary };
        private DictionaryType dictionaryType = DictionaryType.TrieDictionary;
        private const int MAX_WORD_LENGTH = 15;
        private const int MAX_NAME_LENGTH = 100;
        private const int MAX_FILENAME_LENGTH = 100;
        private const string DICTIONARIES_FOLDER = @".\dictionaries\";
        private const string PREPARED_EXTENSION = ".Prepared";

        /*Variables d'instancia*/

        //Classe per accedir al sistema LINQ que connecta amb la base de dades
        private ScrabbleClassesDataContext dataContext;

        //Llista de diccionaris carregats en memoria
        private Dictionary<string, IScrabbleDictionary> dictionaries = new Dictionary<string, IScrabbleDictionary>();

        /* Funció que comproba si per un arxiu de diccionari determinat hi ha el seu corresponent preparat.
         * En cas negatiu el genera en aquest moment.
         * Un arxiu preparat es el mateix diccionari però assegurant les següents condicions:
         * 1º- No conté paraules repetides
         * 2º- Tots els caràcters estan en majuscules
         */
        private string CheckAndPrepareDictionary(string fileName)
        {
            try
            {
                string dictionaryFile = DICTIONARIES_FOLDER + fileName;
                string preparedFile = dictionaryFile + PREPARED_EXTENSION;

                //Si l'arxiu ja existeix no hem de fer res
                if (File.Exists(preparedFile))
                {
                    return preparedFile;
                }

                //Llegim l'arxiu d'entrada i desem les paraules que ens interessin
                Dictionary<string, bool> wordList = new Dictionary<string, bool>();
                using (StreamReader sr = new StreamReader(dictionaryFile))
                {
                    string inWord;

                    while ((inWord = sr.ReadLine()) != null)
                    {
                        inWord = inWord.ToUpper().Trim();

                        if (!(wordList.ContainsKey(inWord)))
                        {
                            wordList.Add(inWord, true);
                        }
                    }
                }

                //Finalment desem la llista preparada
                using (StreamWriter sw = new StreamWriter(preparedFile))
                {
                    foreach (string word in wordList.Keys)
                    {
                        sw.WriteLine(word);
                    }
                }

                return preparedFile;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint al sistema d'arxius:\n" + ex.Message, ScrabbleError.FILESYSTEM_UNEXPECTED_ERROR);
            }
        }

        public IScrabbleDictionary GetDictionary(string dictionaryId)
        {
            IScrabbleDictionary result = null;

            //Primer mirem si ja el tenim carregat en memòria
            if (dictionaries.TryGetValue(dictionaryId, out result))
            {
                return result;
            }
            //sino el carreguem i l'afegim a la llista de diccionaris en memoria
            else
            {
                DICTIONARY dbDictionary = null;

                try
                {
                    var dbDictionaries = (from d in dataContext.DICTIONARY
                                          where d.id == dictionaryId
                                          select d);

                    if (dbDictionaries.Count() > 0)
                    {
                        dbDictionary = dbDictionaries.First();
                    }

                }
                catch (Exception ex)
                {
                    throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
                }


                if (dbDictionary != null)
                {

                    string preparedFile = CheckAndPrepareDictionary(dbDictionary.fileName);

                    switch (dictionaryType)
                    {
                        case DictionaryType.BasicDictionary:
                            {
                                result = new BasicDictionary() { Id = dictionaryId, Name = dbDictionary.name, FileName = preparedFile };
                                break;
                            }
                        case DictionaryType.TrieDictionary:
                            {
                                result = new TrieDictionary { Id = dictionaryId, Name = dbDictionary.name, FileName = preparedFile };
                                break;
                            }
                        default:
                            {
                                throw ScrabbleError.GetNewFaultException("Tipus de diccionari no valid", ScrabbleError.DICTIONARY_HAS_INVALID_TYPE);
                            }
                    }

                    result.Initialize();
                    dictionaries.Add(dictionaryId, result);

                    return result;
                }
                else
                {
                    throw ScrabbleError.GetNewFaultException("No s'ha trobat cap diccionari amb aquest id", ScrabbleError.DICTIONARY_NOT_FOUND);
                }
            }
        }


        /*
         * Métodes per gestionar els diccionaris a través dels clients (es reben i transmeten les classes ...DTO que encapsulen les dades necessaries) 
         */
        public List<DictionaryInfoDTO> GetDictionaryInfos()
        {
            try
            {
                List<DictionaryInfoDTO> result = new List<DictionaryInfoDTO>();

                var dbDictionaries = (from d in dataContext.DICTIONARY
                                      orderby d.name
                                      select d);

                foreach (DICTIONARY d in dbDictionaries)
                {
                    DictionaryInfoDTO dictionaryInfo = new DictionaryInfoDTO() { Id = d.id, Name = d.name, FileName = d.fileName };
                    result.Add(dictionaryInfo);
                }
                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void SaveDictionaryInfo(DictionaryInfoDTO dictionaryInfo)
        {
            if (dictionaryInfo.Name.Length > MAX_NAME_LENGTH)
            {
                throw ScrabbleError.GetNewFaultException(String.Format("No es pot desar la informació de diccionari, el nom de diccionari no pot superar els {0} caracters", MAX_NAME_LENGTH), ScrabbleError.DICTIONARY_NAME_TOO_LONG);
            }

            if (dictionaryInfo.FileName.Length > MAX_FILENAME_LENGTH)
            {
                throw ScrabbleError.GetNewFaultException(String.Format("No es pot desar la informació de diccionari, el nom de l'arxiu no pot superar els {0} caracters", MAX_NAME_LENGTH), ScrabbleError.DICTIONARY_FILENAME_TOO_LONG);
            }

            bool newDictionaryInfo = false;

            //Si l'id està buit es un nou diccionari i assignem un nou id
            if (dictionaryInfo.Id == "")
            {
                dictionaryInfo.Id = Guid.NewGuid().ToString();
                newDictionaryInfo = true;
            }

            //Abans de desar s'han de fer els checks pertinents
            int count;

            try
            {
                //1º el nom no pot estar repetit
                count = (from d in dataContext.DICTIONARY
                         where (d.name == dictionaryInfo.Name && d.id != dictionaryInfo.Id)
                         select d).Count(); ;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("No es pot desar la informació de diccionari, ja hi ha un diccionari amb aquest nom.", ScrabbleError.DICTIONARY_NAME_ALREADY_EXISTS);
            }

            if (newDictionaryInfo)
            {
                SaveNewDictionaryInfo(dictionaryInfo);
            }
            else
            {
                SaveUpdatedDictionaryInfo(dictionaryInfo);
            }

        }

        private void SaveNewDictionaryInfo(DictionaryInfoDTO dictionaryInfo)
        {
            try
            {
                //Inserim la informació de diccionari
                DICTIONARY dictionary = new DICTIONARY();
                dictionary.id = dictionaryInfo.Id;
                dictionary.name = dictionaryInfo.Name;
                dictionary.fileName = dictionaryInfo.FileName;

                dataContext.DICTIONARY.InsertOnSubmit(dictionary);
                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void SaveUpdatedDictionaryInfo(DictionaryInfoDTO dictionaryInfo)
        {
            //No modificar si està en ús
            CheckIfInUse(dictionaryInfo);

            try
            {
                //Actualitzem la informació de diccionari
                var dbDictionary = (from d in dataContext.DICTIONARY
                                    where d.id == dictionaryInfo.Id
                                    select d).First();

                dbDictionary.name = dictionaryInfo.Name;
                dbDictionary.fileName = dictionaryInfo.FileName;
                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void DeleteDictionaryInfo(DictionaryInfoDTO dictionaryInfo)
        {
            //No borrar si està en ús
            CheckIfInUse(dictionaryInfo);

            try
            {
                //Borrem la informació de diccionari
                var dbDictionaries = (from d in dataContext.DICTIONARY
                                      where d.id == dictionaryInfo.Id
                                      select d);

                dataContext.DICTIONARY.DeleteAllOnSubmit(dbDictionaries);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void CheckIfInUse(DictionaryInfoDTO dictionaryInfo)
        {
            int count;

            try
            {
                count = (from g in dataContext.GAME
                         where (g.DICTIONARY1.id == dictionaryInfo.Id) && (g.finished == false)
                         select g).Count();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("Operació no vàlida, el diccionari està en ús", ScrabbleError.DICTIONARY_IS_IN_USE);
            }
        }
    }
}
