﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;
using System.Threading.Tasks;
using System.ServiceModel;
using ServerLibrary.LINQ.Context;
using ServerLibrary.LINQ.Entity;
using ServerLibrary.Domain.Users;
using ServerLibrary.Domain.Games;
using ServerLibrary.Domain.Boards;
using ServerLibrary.Domain.TileSets;
using ServerLibrary.Domain.Dictionaries;
using ServerLibrary.Domain.Errors;
using ServerLibrary.WCF.DTO;

namespace ServerLibrary.LINQ
{
    /*
     * Classe que s'encarrega de la persistencia i creacio de les partides.
     */
    public sealed class GamesFactory
    {
        /*Gestió del singleton*/
        private static readonly GamesFactory instance = new GamesFactory();

        public static GamesFactory Instance
        {
            get
            {
                return instance;
            }
        }

        private GamesFactory()
        {
            try
            {
                string connStr = ConfigurationManager.ConnectionStrings["ServerLibrary.Properties.Settings.SCRABBLEConnectionString"].ConnectionString;
                dataContext = new ScrabbleClassesDataContext(connStr);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        /*Constants*/

        /*Variables d'instancia*/

        //Classe per accedir al sistema LINQ que connecta amb la base de dades
        private ScrabbleClassesDataContext dataContext;

        //Mètode per crear (i desar) una nova partida per un jugador
        public Game GetNewOnePlayerGame(GameInfoDTO gameInfo, User user)
        {
            //Amb les dades que tenim, creem tots els objectes associats a una partida            
            IPlayer player1 = new HumanPlayer(user, user.Id, user.Name);        
            IPlayer player2 = new CPUPlayer();
            
            Board board = BoardsFactory.Instance.GetBoard(gameInfo.Board);
            TileSet tileSet = TileSetsFactory.Instance.GetTileSet(gameInfo.TileSet);            
            IScrabbleDictionary dictionary = DictionariesFactory.Instance.GetDictionary(gameInfo.Dictionary);           

            Game newGame = new Game(Guid.NewGuid().ToString(), gameInfo.Name, player1, player2, board, tileSet, dictionary);
            player1.Game = newGame;
            player2.Game = newGame;

            //Desar la partida a la bbdd
            SaveNewGame(newGame, 1);

            return newGame;
        }

        //Mètode per crear (i desar) una nova partida per dos jugadors
        public Game GetNewTwoPlayerGame(GameInfoDTO gameInfo, User user)
        {
            //Amb les dades que tenim, creem tots els objectes associats a una partida
            IPlayer player1 = new HumanPlayer(user, user.Id, user.Name);
            Board board = BoardsFactory.Instance.GetBoard(gameInfo.Board);
            TileSet tileSet = TileSetsFactory.Instance.GetTileSet(gameInfo.TileSet);
            IScrabbleDictionary dictionary = DictionariesFactory.Instance.GetDictionary(gameInfo.Dictionary);

            Game newGame = new Game(Guid.NewGuid().ToString(), gameInfo.Name, player1, board, tileSet, dictionary);
            player1.Game = newGame;

            //Desar la partida a la bbdd
            SaveNewGame(newGame, 2);

            return newGame;
        }

        //Mètode per esborrar una partida
        public void DeleteGame(Game game)
        {
            GAME dbGame = null;

            try
            {
                var dbGames = (from g in dataContext.GAME
                               where g.id == game.Id
                               select g);

                if (dbGames.Count() > 0)
                {
                    dbGame = dbGames.First();
                }

                //La base de dades esta preparada per fer "Delete on cascade" de les taules relacionades
                dataContext.GAME.DeleteOnSubmit(dbGame);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        //Mètode per esborrar totes les partides on participa un usuari
        public void DeleteAllGamesFromUser(string userId)
        {
            try
            {
                var dbGames = (from g in dataContext.GAME
                               where g.player1 == userId || g.player2 == userId
                               select g);

                //La base de dades esta preparada per fer "Delete on cascade" de les taules relacionades
                dataContext.GAME.DeleteAllOnSubmit(dbGames);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void SaveNewGame(Game newGame, int numPlayers)
        {
            try
            {
                //Creem i omplim l'objecte de bbdd
                GAME dbGame = new GAME();
                dbGame.id = newGame.Id;
                dbGame.name = newGame.Name;
                dbGame.numPlayers = numPlayers;
                dbGame.player1 = newGame.Player1.Id;
                dbGame.player1Score = 0;
                if (numPlayers == 1)
                {
                    dbGame.player2 = newGame.Player2.Id; //Serà la CPU
                }
                dbGame.player2Score = 0;
                dbGame.turn = newGame.Turn;
                dbGame.board = newGame.Board.Id;
                dbGame.tileSet = newGame.TileSet.Id;
                dbGame.dictionary = newGame.Dictionary.Id;
                dbGame.finished = false;

                dataContext.GAME.InsertOnSubmit(dbGame);
                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void SavePlayer2(Game game)
        {
            try
            {
                //Obtenim la partida a actualitzar
                GAME dbGame = dataContext.GAME.Where(g => g.id == game.Id).First();

                dbGame.USER1 = dataContext.USER.Single(u => u.id == game.Player2.Id);
                //dbGame.player2 = game.Player2.Id;

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public Game LoadGame(string gameId)
        {
            //Carreguem una partida de la base de dades
            GAME dbGame = null;

            try
            {
                var dbGames = (from g in dataContext.GAME
                               where g.id == gameId
                               select g);

                if (dbGames.Count() > 0)
                {
                    dbGame = dbGames.First();
                }
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (dbGame != null)
            {
                //Amb les dades que tenim, creem tots els objectes associats a una partida
                string id = dbGame.id;
                string name = dbGame.name;
                int numPlayers = dbGame.numPlayers;
                IPlayer player1 = new HumanPlayer(null, dbGame.USER.id, dbGame.USER.name);
                player1.Score = dbGame.player1Score;
                IPlayer player2 = null;
                //Si la partida es d'un jugador, el player dos es CPUPlayer
                if (numPlayers == 1)
                {
                    player2 = new CPUPlayer();
                    player2.Score = dbGame.player2Score;
                }
                //Si es de dos jugadors potser ja tenia player2 o potser no però serà HumanPlayer.
                else
                {
                    if (dbGame.player2 != null)
                    {
                        player2 = new HumanPlayer(null, dbGame.USER1.id, dbGame.USER1.name);
                        player2.Score = dbGame.player2Score;
                    }
                }
                int turn = dbGame.turn;
                bool started = dbGame.started;
                bool finished = dbGame.finished;
                int turnsWithoutScore = dbGame.turnsWithoutScore;
                Board board = BoardsFactory.Instance.GetBoard(dbGame.board);
                TileSet tileSet = TileSetsFactory.Instance.GetTileSet(dbGame.tileSet);             
                IScrabbleDictionary dictionary = DictionariesFactory.Instance.GetDictionary(dbGame.dictionary);

                Game game = new Game(id, name, player1, player2, board, tileSet, dictionary, turn, started, finished, turnsWithoutScore);
                player1.Game = game;
                if (player2 != null)
                {
                    player2.Game = game;
                }

                //Restaurem els moviments
                List<Move> moves = new List<Move>();
                foreach (MOVE m in dbGame.MOVE.OrderBy(m => m.moveNumber))
                {
                    Move move = new Move();
                    if (player1.Id == m.player)
                    {
                        move.Player = player1;
                        player1.Score += move.Score;
                    }
                    else if (player2 != null && player2.Id == m.player)
                    {
                        move.Player = player2;
                        player2.Score += move.Score;
                    }

                    move.InitialRow = m.initialRow;
                    move.InitialColumn = m.initialColumn;
                    move.Direction = (Move.MoveDirection)m.direction;
                    move.Score = m.score;

                    moves.Add(move);

                    List<MoveTile> moveTiles = new List<MoveTile>();
                    foreach (MOVE_TILE mt in m.MOVE_TILE.OrderBy(mt => mt.tileNumber))
                    {
                        Tile tile = tileSet.GetTile(mt.tile);
                        //Si no hem recuperat la fitxa es perque ja l'hem extret en una jugada anterior
                        if (tile != null)
                        {
                            //Si la fitxa es un escarràs, en recuperem la lletra amb el que es va jugar
                            if (tile.Type == Tile.TileType.Blank)
                            {
                                tile.Letter = mt.letter;
                            }
                        }

                        MoveTile moveTile = new MoveTile();
                        moveTile.Tile = tile;
                        moveTile.Row = mt.tileRow;
                        moveTile.Column = mt.tileColumn;
                        moveTile.NewTile = mt.newTile;

                        moveTiles.Add(moveTile);
                    }

                    move.Tiles = moveTiles;
                }

                game.Moves = moves;               

                //Restaurem els faristols
                LoadPlayerRack(game, player1, tileSet);
                if (player2 != null)
                {
                    LoadPlayerRack(game, player2, tileSet);
                }

                //Barrejem les fitxes que quedin
                tileSet.Shuffle();

                //Si es una partida d'un jugador i el torn es per la CPU, l'avisem perque comenci a processar la jugada
                if (numPlayers == 1 && turn == 1)
                {
                    player2.YourTurn();
                }

                return game;
            }
            else
            {
                throw ScrabbleError.GetNewFaultException("No s'ha trobat cap partida amb aquest id: " + gameId, ScrabbleError.GAME_NOT_FOUND);
            }
        }

        private void LoadPlayerRack(Game game, IPlayer player, TileSet tileSet)
        {
            try
            {
                Rack rack;
                RACK dbRack;

                //Jugador 1
                var dbRacks = (from r in dataContext.RACK
                               where r.game == game.Id && r.player == player.Id
                               select r);

                if (dbRacks.Count() > 0)
                {
                    dbRack = dbRacks.First();

                    rack = new Rack();
                    rack.SetTile(0, tileSet.GetTile(dbRack.p1));
                    rack.SetTile(1, tileSet.GetTile(dbRack.p2));
                    rack.SetTile(2, tileSet.GetTile(dbRack.p3));
                    rack.SetTile(3, tileSet.GetTile(dbRack.p4));
                    rack.SetTile(4, tileSet.GetTile(dbRack.p5));
                    rack.SetTile(5, tileSet.GetTile(dbRack.p6));
                    rack.SetTile(6, tileSet.GetTile(dbRack.p7));

                    player.Rack = rack;
                }
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void UpdateGame(Game game)
        {
            try
            {
                //Obtenim la partida a actualitzar
                GAME dbGame = dataContext.GAME.Where(g => g.id == game.Id).First();

                dbGame.started = game.Started;
                dbGame.finished = game.Finished;
                dbGame.turn = game.Turn;
                dbGame.turnsWithoutScore = game.TurnsWithoutScore;
                dbGame.player1Score = game.Player1.Score;
                dbGame.player2Score = game.Player2.Score;

                SavePlayerRack(game, game.Player1);
                SavePlayerRack(game, game.Player2);

                //Si la partida ha acabat, s'ha de desar estadístiques 
                if (game.Finished)
                {
                    if (game.Player1.Score > game.Player2.Score)
                    {
                        UpdatePlayerGameStats(game.Player1, 0);
                        UpdatePlayerGameStats(game.Player2, 2);
                    }
                    else if (game.Player1.Score < game.Player2.Score)
                    {
                        UpdatePlayerGameStats(game.Player1, 2);
                        UpdatePlayerGameStats(game.Player2, 0);
                    }
                    else
                    {
                        UpdatePlayerGameStats(game.Player1, 1);
                        UpdatePlayerGameStats(game.Player2, 1);
                    }
                }

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

        }

        private void UpdatePlayerGameStats(IPlayer player, int gameResult)
        {
            //Desem les dades de partides guanyades, perdudes...
            STATS_GAME stGame;

            //Busquem el jugador a la taula corresponent
            var stGames = (from stg in dataContext.STATS_GAME
                           where stg.playerId == player.Id
                           select stg);

            if (stGames.Count() > 0)
            {
                stGame = stGames.First();
            }
            //Si no hi es el creem nou
            else
            {
                stGame = new STATS_GAME();
                stGame.playerId = player.Id;
                stGame.playerName = player.Name;
                dataContext.STATS_GAME.InsertOnSubmit(stGame);
            }

            switch (gameResult)
            {
                case 0:
                    stGame.winGames++;
                    break;

                case 1:
                    stGame.drawGames++;
                    break;

                case 2:
                    stGame.lostGames++;
                    break;
            }

            //I la puntuació de la partida que ha aconseguit el jugador
            STATS_GAME_SCORE stGameScore = new STATS_GAME_SCORE();
            stGameScore.playerId = player.Id;
            stGameScore.playerName = player.Name;
            stGameScore.gameScore = player.Score;
            stGameScore.gameDateTime = DateTime.UtcNow;

            dataContext.STATS_GAME_SCORE.InsertOnSubmit(stGameScore);
        }

        public void SaveMove(Game game)
        {
            //Quan es guarda un moviment s'ha de desar l'ultim moviment de la partida
            // i el faristol del jugador que l'ha realitzat
            try
            {
                //Obtenim la partida a actualitzar
                GAME dbGame = dataContext.GAME.Where(g => g.id == game.Id).First();

                //Creem el nou moviment
                MOVE dbMove = new MOVE();
                Move lastMove = game.Moves[game.Moves.Count() - 1];

                dbMove.id = Guid.NewGuid().ToString();
                dbMove.game = game.Id;
                dbMove.moveNumber = game.Moves.Count();

                dbMove.player = game.TurnPlayer.Id;
                dbMove.initialRow = lastMove.InitialRow;
                dbMove.initialColumn = lastMove.InitialColumn;
                dbMove.direction = (int)lastMove.Direction;
                dbMove.score = lastMove.Score;

                dataContext.MOVE.InsertOnSubmit(dbMove);

                MOVE_TILE moveTile;

                //Per cada fitxa de la jugada creem una nova entrada a la taula corresponent
                for (int i = 0; i < lastMove.Tiles.Count(); i++)
                {
                    MoveTile mt = lastMove.Tiles[i];
                    Tile t = mt.Tile;
                    moveTile = new MOVE_TILE();

                    moveTile.MOVE1 = dbMove;
                    moveTile.tileNumber = i;
                    moveTile.tileRow = mt.Row;
                    moveTile.tileColumn = mt.Column;
                    moveTile.tile = t.Id;
                    //Afegim la lletra perque la fitxa pot ser un escarras i no tenir lletra assignada fins que no es fa la jugada
                    //Si no la desem, despres no podrem recuperar la lletra amb la que es va jugar l'escarràs
                    moveTile.letter = t.Letter;
                    moveTile.newTile = mt.NewTile;

                    dataContext.MOVE_TILE.InsertOnSubmit(moveTile);
                }


                //Desem en un altre taula el resum del moviment per treure estadistiques
                STATS_MOVE stMove = new STATS_MOVE();
                stMove.playerId = game.TurnPlayer.Id;
                stMove.playerName = game.TurnPlayer.Name;
                stMove.moveScore = lastMove.Score;
                stMove.moveWords = string.Join<string>(", ", lastMove.Words);
                stMove.moveDateTime = DateTime.UtcNow;

                dataContext.STATS_MOVE.InsertOnSubmit(stMove);

                //no fem el submit aquí, es farà quan es cridi a change turn (sino no es transaccional)
                //dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void SavePlayerRack(Game game, IPlayer player)
        {
            //Mirem si el faristol ja està creat
            RACK dbRack;

            var dbRacks = (from r in dataContext.RACK
                           where r.game == game.Id && r.player == player.Id
                           select r);

            if (dbRacks.Count() > 0)
            {
                dbRack = dbRacks.First();
            }
            else
            {
                dbRack = new RACK();
                dbRack.game = game.Id;
                dbRack.player = player.Id;
                dataContext.RACK.InsertOnSubmit(dbRack);
            }

            FillRack(dbRack, player.Rack);
        }

        private void FillRack(RACK rack, Rack playerRack)
        {
            rack.p1 = playerRack.GetTileWithoutRemove(0) != null ? playerRack.GetTileWithoutRemove(0).Id : null ;
            rack.p2 = playerRack.GetTileWithoutRemove(1) != null ? playerRack.GetTileWithoutRemove(1).Id : null;
            rack.p3 = playerRack.GetTileWithoutRemove(2) != null ? playerRack.GetTileWithoutRemove(2).Id : null;
            rack.p4 = playerRack.GetTileWithoutRemove(3) != null ? playerRack.GetTileWithoutRemove(3).Id : null;
            rack.p5 = playerRack.GetTileWithoutRemove(4) != null ? playerRack.GetTileWithoutRemove(4).Id : null;
            rack.p6 = playerRack.GetTileWithoutRemove(5) != null ? playerRack.GetTileWithoutRemove(5).Id : null;
            rack.p7 = playerRack.GetTileWithoutRemove(6) != null ? playerRack.GetTileWithoutRemove(6).Id : null;
        }

        /*
         * Métodes per gestionar les partides a través dels clients (es reben i transmeten les classes ...DTO que encapsulen les dades necessaries) 
         */

        //Mètode que retornarà la llista de partides d'altres jugadors pendents de trobar un segon jugador
        public List<GameInfoDTO> GetAvailableGames(string userId)
        {
            try
            {
                List<GameInfoDTO> result = new List<GameInfoDTO>();

                var dbGames = (from g in dataContext.GAME
                               where (g.numPlayers == 2) && (g.started == false) && (g.player2 == null) && (g.player1 != userId)
                               orderby g.name
                               select g);

                foreach (GAME dbGame in dbGames)
                {
                    GameInfoDTO gameInfo = new GameInfoDTO();
                    gameInfo.Id = dbGame.id;
                    gameInfo.Name = dbGame.name;

                    gameInfo.BoardName = dbGame.BOARD1.name;
                    gameInfo.Board = dbGame.BOARD1.id;
                    gameInfo.TileSetName = dbGame.TILESET1.name;
                    gameInfo.TileSet = dbGame.TILESET1.id;
                    gameInfo.DictionaryName = dbGame.DICTIONARY1.name;
                    gameInfo.Dictionary = dbGame.DICTIONARY1.id;

                    gameInfo.Player1Name = dbGame.USER.name;
                    gameInfo.Player1 = dbGame.USER.id;

                    result.Add(gameInfo);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        //Mètode que retornarà la llista de partides en que participa l'usuari
        public List<GameInfoDTO> GetCurrentUserGames(string userId)
        {
            try
            {
                List<GameInfoDTO> result = new List<GameInfoDTO>();

                var dbGames = (from g in dataContext.GAME
                               where g.player1 == userId || g.player2 == userId
                               orderby g.name
                               select g);

                foreach (GAME dbGame in dbGames)
                {
                    GameInfoDTO gameInfo = new GameInfoDTO();
                    gameInfo.Id = dbGame.id;
                    gameInfo.Name = dbGame.name;

                    gameInfo.BoardName = dbGame.BOARD1.name;
                    gameInfo.Board = dbGame.BOARD1.id;
                    gameInfo.TileSetName = dbGame.TILESET1.name;
                    gameInfo.TileSet = dbGame.TILESET1.id;
                    gameInfo.DictionaryName = dbGame.DICTIONARY1.name;
                    gameInfo.Dictionary = dbGame.DICTIONARY1.id;
                    gameInfo.Turn = dbGame.turn;

                    gameInfo.Player1Name = dbGame.USER.name;
                    gameInfo.Player1 = dbGame.USER.id;
                    gameInfo.Player1Score = dbGame.player1Score;

                    if (dbGame.numPlayers == 2)
                    {
                        if (dbGame.USER1 != null)
                        {
                            gameInfo.Player2Name = dbGame.USER1.name;
                            gameInfo.Player2 = dbGame.USER1.id;
                            gameInfo.Player2Score = dbGame.player2Score;
                        }

                        else
                        {
                            gameInfo.Player2Name = "Encara no hi ha jugador";
                            gameInfo.Player2 = null;
                            gameInfo.Player2Score = 0;
                        }
                    }
                    else
                    {
                        gameInfo.Player2Name = "CPU";
                        gameInfo.Player2 = "";
                        gameInfo.Player2Score = dbGame.player2Score;
                    }
                    gameInfo.Started = dbGame.started;
                    gameInfo.Finished = dbGame.finished;

                    result.Add(gameInfo);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }


        /*
         * Métodes per obtenir les estadístiques relacionades amb les partides (es transmeten les classes ...DTO que encapsulen les dades necessaries) 
         */

        //Torna el Top10 de de millors partides
        public List<StatsHighGameDTO> GetHighestGameScores(bool onlyUserStats, bool excludeCPU, string userId)
        {
            try
            {
                List<StatsHighGameDTO> result = new List<StatsHighGameDTO>();

                //Com que la podem inicialitzar amb diferents queries no admet "var"
                IQueryable<ServerLibrary.LINQ.Entity.STATS_GAME_SCORE> dbStats;

                if (onlyUserStats)
                {
                    dbStats = (from s in dataContext.STATS_GAME_SCORE
                               where s.playerId == userId
                               orderby s.gameScore descending
                               select s).Take(10);
                }
                else if (excludeCPU)
                {
                    dbStats = (from s in dataContext.STATS_GAME_SCORE
                               where s.playerId != CPUPlayer.CPUPLAYERID
                               orderby s.gameScore descending
                               select s).Take(10);
                }
                else
                {
                    dbStats = (from s in dataContext.STATS_GAME_SCORE
                               orderby s.gameScore descending
                               select s).Take(10);
                }

                foreach (STATS_GAME_SCORE dbStat in dbStats)
                {
                    StatsHighGameDTO gameStat = new StatsHighGameDTO();

                    gameStat.PlayerName = dbStat.playerName;
                    gameStat.GameScore = dbStat.gameScore;
                    gameStat.GameDateTime = dbStat.gameDateTime;

                    result.Add(gameStat);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        //Torna el Top10 de de millors jugades
        public List<StatsHighMoveDTO> GetHighestMoveScores(bool onlyUserStats, bool excludeCPU, string userId)
        {
            try
            {
                List<StatsHighMoveDTO> result = new List<StatsHighMoveDTO>();

                //Com que la podem inicialitzar amb diferents queries no admet "var"
                IQueryable<ServerLibrary.LINQ.Entity.STATS_MOVE> dbStats;

                if (onlyUserStats)
                {
                    dbStats = (from s in dataContext.STATS_MOVE
                               where s.playerId == userId
                               orderby s.moveScore descending
                               select s).Take(10);
                }
                else if (excludeCPU)
                {
                    dbStats = (from s in dataContext.STATS_MOVE
                               where s.playerId != CPUPlayer.CPUPLAYERID
                               orderby s.moveScore descending
                               select s).Take(10);
                }
                else
                {
                    dbStats = (from s in dataContext.STATS_MOVE
                               orderby s.moveScore descending
                               select s).Take(10);
                }

                //Si hem modificat el nom d'usuari, sembla que el dataContext ho pot tenir en cache i no mostra el nom actualitzat
                dataContext.Refresh(System.Data.Linq.RefreshMode.OverwriteCurrentValues, dbStats);

                foreach (STATS_MOVE dbStat in dbStats)
                {
                    StatsHighMoveDTO gameStat = new StatsHighMoveDTO();

                    gameStat.PlayerName = dbStat.playerName;
                    gameStat.MoveScore = dbStat.moveScore;
                    gameStat.MoveDateTime = dbStat.moveDateTime;
                    gameStat.MoveWords = dbStat.moveWords;

                    result.Add(gameStat);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        //Torna el Top10 de de millors percentatges de partides guanyades
        public List<StatsGamePercentages> GetHighestWinGamesPercent(bool onlyUserStats, bool excludeCPU, string userId)
        {
            try
            {
                List<StatsGamePercentages> result = new List<StatsGamePercentages>();

                //Com que la podem inicialitzar amb diferents queries no admet "var"
                IQueryable<ServerLibrary.LINQ.Entity.STATS_GAME> dbStats;

                if (onlyUserStats)
                {
                    dbStats = (from s in dataContext.STATS_GAME
                               where s.playerId == userId
                               select s);
                }
                else if (excludeCPU)
                {
                    dbStats = (from s in dataContext.STATS_GAME
                               where s.playerId != CPUPlayer.CPUPLAYERID
                               orderby s.percentWin descending
                               select s).Take(10);
                }
                else
                {
                    dbStats = (from s in dataContext.STATS_GAME
                               orderby s.percentWin descending
                               select s).Take(10);
                }

                foreach (STATS_GAME dbStat in dbStats)
                {
                    StatsGamePercentages gameStat = new StatsGamePercentages();

                    gameStat.PlayerName = dbStat.playerName;
                    gameStat.PercentWin = (dbStat.percentWin != null? (int)dbStat.percentWin : 0);
                    gameStat.PercentDraw = (dbStat.percentDraw != null ? (int)dbStat.percentDraw : 0);
                    gameStat.PercentLost = (dbStat.percentLost != null ? (int)dbStat.percentLost : 0);
                    gameStat.TotalWin = dbStat.winGames;
                    gameStat.TotalDraw = dbStat.drawGames;
                    gameStat.TotalLost = dbStat.lostGames;

                    result.Add(gameStat);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        //Esborra les estadístiques d'un jugador
        public void DeleteUserStats(string userId)
        {
            try
            {
                //Esborrem les estadístiques de partides
                var dbGameStats = (from gs in dataContext.STATS_GAME_SCORE
                                   where gs.playerId == userId
                                   select gs);

                dataContext.STATS_GAME_SCORE.DeleteAllOnSubmit(dbGameStats);

                //Les de jugades
                var dbMoveStats = (from ms in dataContext.STATS_MOVE
                                   where ms.playerId == userId
                                   select ms);

                dataContext.STATS_MOVE.DeleteAllOnSubmit(dbMoveStats);

                //I les de percentatges
                var dbPercentStats = (from ps in dataContext.STATS_GAME
                                   where ps.playerId == userId
                                   select ps);

                dataContext.STATS_GAME.DeleteAllOnSubmit(dbPercentStats);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }
    }
}
