﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;
using System.Threading.Tasks;
using ServerLibrary.WCF.DTO;
using ServerLibrary.Domain.TileSets;
using ServerLibrary.Domain.Users;
using ServerLibrary.Domain.Errors;
using ServerLibrary.LINQ.Context;
using ServerLibrary.LINQ.Entity;
using System.ServiceModel;

namespace ServerLibrary.LINQ
{
    /*
     * Classe que s'encarrega de la gestió i la persistencia dels sacs de fitxes.
     */
    public sealed class TileSetsFactory
    {
        /*Gestió del singleton*/
        private static readonly TileSetsFactory instance = new TileSetsFactory();

        public static TileSetsFactory Instance
        {
            get
            {
                return instance;
            }
        }

        private TileSetsFactory() 
        {
            try
            {
                string connStr = ConfigurationManager.ConnectionStrings["ServerLibrary.Properties.Settings.SCRABBLEConnectionString"].ConnectionString;
                dataContext = new ScrabbleClassesDataContext(connStr);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        /*Constants*/ 
        private const int MAX_TILESET_NAME_LENGTH = 100;

        /*Variables d'instancia*/

        //Classe per accedir al sistema LINQ que connecta amb la base de dades
        private ScrabbleClassesDataContext dataContext;

        /*Mètode que retorna un sac de fitxes amb les seves fitxes*/
        public TileSet GetTileSet(string tileSetId)
        {
            try
            {
                TileSet result = null;

                var dbTileSets = (from ts in dataContext.TILESET
                                  where ts.id == tileSetId
                                  select ts);

                //Si existeix, creem l'objecte i omplim les fitxes
                if (dbTileSets.Count() > 0)
                {
                    TILESET dbTileSet = dbTileSets.First();

                    result = new TileSet() { Id = dbTileSet.id, Name = dbTileSet.name, Date = dbTileSet.creationDate };

                    var dbTiles = (from t in dbTileSet.TILE
                                   select t);

                    foreach (TILE dbTile in dbTiles)
                    {
                        Tile tile = new Tile() { Id = dbTile.id, Letter = dbTile.letter, Value = dbTile.value, Type = (Tile.TileType)dbTile.type };
                        result.AddTile(tile);
                    }
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);              
            }
        }

        //Métodes per gestionar els sacs de fitxes a través dels clients (es reben i transmeten les classes ...DTO que encapsulen les dades necessaries)
        public List<TileSetDTO> GetUserTileSets(string userId)
        {
            try
            {
                List<TileSetDTO> result = new List<TileSetDTO>();

                var dbTileSets = (from t in dataContext.TILESET
                                  where t.owner == userId
                                  select t);

                foreach (TILESET t in dbTileSets)
                {
                    TileSetDTO tileSt = new TileSetDTO() { Owner = t.owner, Id = t.id, Name = t.name, Date = t.creationDate };
                    result.Add(tileSt);
                }
                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public List<TileSetDTO> GetTileSets(string userId)
        {
            try
            {
                List<TileSetDTO> result = new List<TileSetDTO>();

                var dbTileSets = (from t in dataContext.TILESET
                                  where t.owner == userId || t.owner == null
                                  orderby t.owner, t.name
                                  select t);

                foreach (TILESET t in dbTileSets)
                {
                    TileSetDTO tileSet = new TileSetDTO() { Owner = t.owner, Id = t.id, Name = t.name, Date = t.creationDate };
                    result.Add(tileSet);
                }
                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public List<TileDTO> GetTileSetTiles(string tileSetId)
        {
            try
            {
                List<TileDTO> result = new List<TileDTO>();

                var dbTiles = (from t in dataContext.TILE
                               where t.tileSet == tileSetId
                               orderby t.letter ascending
                               select t);

                foreach (TILE dbTile in dbTiles)
                {
                    TileDTO tile = new TileDTO() { Id = dbTile.id, Letter = dbTile.letter.ToString(), Value = dbTile.value, Type = dbTile.type };
                    result.Add(tile);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void SaveTileSet(TileSetDTO tileSet, List<TileDTO> tiles, string userId)
        {
            if (tileSet.Name.Length > MAX_TILESET_NAME_LENGTH)
            {
                throw ScrabbleError.GetNewFaultException(String.Format("No es pot desar el sac de fitxes, el nom no pot superar els {0} caracters", MAX_TILESET_NAME_LENGTH), ScrabbleError.TILESET_NAME_TOO_LONG);
            }

            bool newTileSet = false;

            //Si l'id està buit es un nou sac i assignem un nou id
            if (tileSet.Id == "")
            {
                tileSet.Id = Guid.NewGuid().ToString();
                newTileSet = true;
            }

            //Abans de desar s'han de fer els checks pertinents
            int count = 0;

            try
            {
                //1º el nom no pot estar repetit (dins els sacs del propi usuari)
                count = (from t in dataContext.TILESET
                         where (t.name == tileSet.Name) && (t.owner == userId) && (t.id != tileSet.Id)
                         select t).Count();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("No es pot desar el sac de fitxes, ja hi ha un sac de fitxes amb aquest nom.", ScrabbleError.TILESET_NAME_ALREADY_EXISTS);
            }

            //2º hi ha d'haver exactament 100 fitxes
            if (tiles.Count() != TileSet.NUMBER_OF_TILES)
            {
                throw ScrabbleError.GetNewFaultException("No es pot desar el sac de fitxes, número de fitxes invalid.", ScrabbleError.TILESET_HAS_INVALID_TILES_NUMBER);
            }

            if (newTileSet)
            {
                SaveNewTileSet(tileSet, tiles, userId);
            }
            else
            {
                SaveUpdatedTileSet(tileSet, tiles, userId);
            }
        }

        private void SaveNewTileSet(TileSetDTO tileSet, List<TileDTO> tiles, string userId)
        {
            try
            {
                //Inserim el sac
                TILESET dbTileSet = new TILESET();
                dbTileSet.id = tileSet.Id;
                dbTileSet.name = tileSet.Name;
                dbTileSet.owner = userId;
                dbTileSet.creationDate = DateTime.UtcNow;

                //Inserim les fitxes
                foreach (TileDTO tile in tiles)
                {
                    TILE dbTile = new TILE();
                    dbTile.id = Guid.NewGuid().ToString();
                    dbTile.letter = tile.Letter;
                    dbTile.value = tile.Value;
                    dbTile.type = tile.Type;

                    dbTileSet.TILE.Add(dbTile);
                }

                dataContext.TILESET.InsertOnSubmit(dbTileSet);
                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void SaveUpdatedTileSet(TileSetDTO tileSet, List<TileDTO> tiles, string userId)
        {
            //No es pot modificar si està en us en alguna partida
            CheckIfInUse(tileSet);

            try
            {
                //Actualitzem el sac
                var dbTileSet = (from ts in dataContext.TILESET
                                 where ts.id == tileSet.Id
                                 select ts).First();

                dbTileSet.name = tileSet.Name;

                //Actualitzem les fitxes
                var dbTiles = (from t in dbTileSet.TILE
                               select t);

                foreach (TILE dbTile in dbTiles)
                {
                    TileDTO tile = tiles.Find(t => t.Id == dbTile.id);
                    dbTile.letter = tile.Letter;
                    dbTile.value = tile.Value;
                    dbTile.type = tile.Type;
                }

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void DeleteTileSet(TileSetDTO tileSet, string userId)
        {
            //Quan hi hagi partides no borrar si està en us
            CheckIfInUse(tileSet);

            try
            {
                //Obtenim l'objecte sac
                var dbTileSet = (from ts in dataContext.TILESET
                                 where ts.id == tileSet.Id && ts.owner == userId //Els GUIDs haurien de ser unics, pero forçem a comprobar també el usuari
                                 select ts).First();

                dataContext.TILESET.DeleteOnSubmit(dbTileSet);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        public void DeleteAllUserTileSets(string userId)
        {
            try
            {
                //Obtenim els objectes sac
                var dbTileSets = (from ts in dataContext.TILESET
                                 where ts.owner == userId
                                 select ts);

                dataContext.TILESET.DeleteAllOnSubmit(dbTileSets);

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }

        private void CheckIfInUse(TileSetDTO tileSet)
        {
            int count;

            try
            {
                count = (from g in dataContext.GAME
                         where (g.TILESET1.id == tileSet.Id) && (g.finished == false)
                         select g).Count();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("Operació no vàlida, el sac de fitxes està en ús", ScrabbleError.TILESET_IS_IN_USE);
            }
        }
    }
}
