﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;
using System.Threading.Tasks;
using System.ServiceModel;
using ServerLibrary.LINQ.Context;
using ServerLibrary.LINQ.Entity;
using ServerLibrary.Domain.Users;
using ServerLibrary.WCF.DTO;
using ServerLibrary.Domain.Errors;

namespace ServerLibrary.LINQ
{
    public sealed class UsersFactory
    {
        private const int MAX_USER_NAME_LENGTH = 100;
        private static readonly UsersFactory instance = new UsersFactory();

        public static UsersFactory Instance
        {
            get
            {
                return instance;
            }
        }

        private UsersFactory() 
        {
            try
            {
                string connStr = ConfigurationManager.ConnectionStrings["ServerLibrary.Properties.Settings.SCRABBLEConnectionString"].ConnectionString;
                dataContext = new ScrabbleClassesDataContext(connStr);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        private ScrabbleClassesDataContext dataContext;

        public User LoadUser(string name)
        {
            //A partir del nom carreguem la resta de la informació
            USER dbUser = null;
            try
            {
                var dbUsers = (from u in dataContext.USER
                            where u.name == name
                            select u);

                if (dbUsers.Count() > 0)
                {
                    dbUser = dbUsers.First();
                }
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (dbUser != null)
            {
                User result = new User() { Id = dbUser.id, Name = dbUser.name, Password = dbUser.password, IsAdmin = (dbUser.isAdmin != null? (bool)dbUser.isAdmin : false) };
                return result;
            }
            else
            {
                throw ScrabbleError.GetNewFaultException("No s'ha trobat cap usuari amb aquest nom: " + name, ScrabbleError.USER_NOT_FOUND);
            }
        }


        public UserDTO GetUserInfo(User user)
        {
            return new UserDTO() { Id = user.Id, Name = user.Name, IsAdmin = user.IsAdmin };
        }


        public void SaveNewUser(UserDTO newUser)
        {
            if (newUser.Name.Length > MAX_USER_NAME_LENGTH)
            {
                throw ScrabbleError.GetNewFaultException(String.Format("No es pot crear l'usuari, el nom no pot superar els {0} caracters", MAX_USER_NAME_LENGTH), ScrabbleError.USER_NAME_TOO_LONG);
            }

            newUser.Id = Guid.NewGuid().ToString();

            //Comprovem el nom d'usuari a la base de dades
            int count = 0;

            try
            {
                count = (from u in dataContext.USER
                         where u.name == newUser.Name && u.id != newUser.Id
                         select u).Count();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("No es pot crear l'usuari, ja hi ha un usuari amb aquest nom.", ScrabbleError.USER_NAME_ALREADY_EXISTS);
            }
            
            try
            {
                USER dbUser = new USER { id = newUser.Id, name = newUser.Name, password = newUser.Password, isAdmin = false };
                dataContext.USER.InsertOnSubmit(dbUser);
                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }


        public void UpdateUser(User updatedUser)
        {
            if (updatedUser.Name.Length > MAX_USER_NAME_LENGTH)
            {
                throw ScrabbleError.GetNewFaultException(String.Format("No es pot desar l'usuari, el nom no pot superar els {0} caracters", MAX_USER_NAME_LENGTH), ScrabbleError.USER_NAME_TOO_LONG);
            }

            //Comprovem el nom d'usuari a la base de dades
            int count = 0;

            try
            {
                count = (from u in dataContext.USER
                         where u.name == updatedUser.Name && u.id != updatedUser.Id
                         select u).Count();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }

            if (count > 0)
            {
                throw ScrabbleError.GetNewFaultException("No es pot desar l'usuari, el nom ja existeix.", ScrabbleError.USER_NAME_ALREADY_EXISTS);
            }

            try
            {
                bool nameChanged = false;

                var dbUser = (from u in dataContext.USER
                              where u.id == updatedUser.Id
                              select u).First();

                nameChanged = dbUser.name != updatedUser.Name;

                dbUser.name = updatedUser.Name;
                dbUser.password = updatedUser.Password;
                

                //Actualitzem les taules d'estadístiques amb el nom nou
                /*Lamentablement LINQ no permet fer updates de cop en un conjunt de registres, s'han de recuperar, iterar i modificar
                 Seria mes optim realitzar directament una sentencia SQL de tipus update contra la bbdd, pero per aquest projecte
                 hem de fer servir LINQ i en aquest cas això prima sobre l'eficiencia*/
                //if (nameChanged)
                {
                    var dbGameStats = (from gs in dataContext.STATS_GAME_SCORE
                                       where gs.playerId == updatedUser.Id
                                       select gs);

                    foreach (STATS_GAME_SCORE dbGameStat in dbGameStats)
                    {
                        dbGameStat.playerName = updatedUser.Name;
                    }

                    var dbMoveStats = (from ms in dataContext.STATS_MOVE
                                       where ms.playerId == updatedUser.Id
                                       select ms);

                    foreach (STATS_MOVE dbMoveStat in dbMoveStats)
                    {
                        dbMoveStat.playerName = updatedUser.Name;
                    }

                    var dbPercentStats = (from ms in dataContext.STATS_GAME
                                       where ms.playerId == updatedUser.Id
                                       select ms);

                    foreach (STATS_GAME dbPercentStat in dbPercentStats)
                    {
                        dbPercentStat.playerName = updatedUser.Name;
                    }
                }

                dataContext.SubmitChanges();
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }


        public void DeleteUser(User removedUser, bool deleteStats)
        {

            //Esborrar tot el que depen de l'usuari (estadistiques, taulers, sacs...)
            TileSetsFactory.Instance.DeleteAllUserTileSets(removedUser.Id);
            BoardsFactory.Instance.DeleteAllUserBoards(removedUser.Id);

            try
            {           
                var dbUser = from u in dataContext.USER
                             where u.id == removedUser.Id
                             select u;

                if (dbUser.Count() > 0)
                {
                    dataContext.USER.DeleteOnSubmit(dbUser.First());
                    dataContext.SubmitChanges();
                }

                //Si hem de borrar estadistíques cridem al mètode corresponent
                if (deleteStats)
                {
                    GamesFactory.Instance.DeleteUserStats(removedUser.Id);
                }
            }
            catch (Exception ex)
            {
                throw ScrabbleError.GetNewFaultException("Error inesperat accedint a base de dades:\n" + ex.Message, ScrabbleError.DATABASE_UNEXPECTED_ERROR);
            }
        }
    }
}
