﻿using System;
using System.Collections.Generic;
using System.ServiceModel;
using ServerLibrary.Domain.Games;
using ServerLibrary.Domain.Users;
using ServerLibrary.Domain.Errors;
using ServerLibrary.LINQ;
using ServerLibrary.WCF.DTO;

namespace ServerLibrary
{
    [ServiceBehavior(ConcurrencyMode = ConcurrencyMode.Reentrant,
                     InstanceContextMode = InstanceContextMode.PerSession)]
    public class ScrabbleService : IScrabbleService
    {
        private User user;


        public void HeartBeat()
        {
            //Metode que cridara el client periodicament per tal que no caduqui la sessió mentre el client estigui corrent i l'usuari identificat
            Console.WriteLine(user.Name + ": HeartBeat");
            //No fer res, serveix nomes per mantenir la sessio oberta
        }


        /*
         * *************************************************************************************
         * GESTIÓ D'USUARIS
         * *************************************************************************************
         */

        //Crea un nou usuari
        public void CreateUser(UserDTO userDTO)
        {
            Console.WriteLine("CreateUser: " + userDTO.Name);

            if (userDTO.Name == "" || userDTO.Password == "")
            {
                throw ScrabbleError.GetNewFaultException("El nom d'usuari o el password no poden estar en blanc", ScrabbleError.USER_OR_PASSWORD_BLANK);
            }

            UsersFactory.Instance.SaveNewUser(userDTO);
        }

        //Inicia sessió i es desa una referencia al callback
        public UserDTO LogIn()
        { 
            Console.WriteLine(ServiceSecurityContext.Current.PrimaryIdentity.Name + ": LogIn");

            user = UsersFactory.Instance.LoadUser(ServiceSecurityContext.Current.PrimaryIdentity.Name);
            user.ClientCallback = OperationContext.Current.GetCallbackChannel<IScrabbleServiceCallback>();

            return UsersFactory.Instance.GetUserInfo(user);
        }

        //Finalitza la sessió
        public void LogOut()
        {
            Console.WriteLine(user.Name + ": LogOut");

            //Es desvincula de les partides carregades en memòria
            GameManager.Instance.DetachFromAllGames(user);
        }

        //Actualitza les dades de l'usuari
        public void UpdateUser(UserDTO userInfo)
        {
            Console.WriteLine(user.Name + ": UpdateUser");

            user.Name = userInfo.Name;
            user.Password = userInfo.Password;
            UsersFactory.Instance.UpdateUser(user);
        }

        //Elimina l'usuari (i les seves estadístiques si s'escau)
        public void DeleteUser(bool deleteStats)
        {
            Console.WriteLine(user.Name + ": DeleteUser");

            //Primer hem d'eliminar totes les partides en que participa l'usuari
            GameManager.Instance.DeleteAllGamesFromUser(user);

            //I despres tota la resta
            UsersFactory.Instance.DeleteUser(user, deleteStats);
        }

        private void EnsureUserIsAdmin()
        {
            if (!user.IsAdmin)
            {
                throw ScrabbleError.GetNewFaultException("Aquesta operació nomes la pot realitzar un administrador", ScrabbleError.USER_HAS_NOT_REQUIRED_PRIVILEGE);
            }
        }


        /*
         * *************************************************************************************
         * GESTIÓ DE TAULERS
         * *************************************************************************************
         */

        //Retorna la llista de taulers de l'usuari identificat
        public List<BoardDTO> GetUserBoards()
        {
            Console.WriteLine(user.Name + ": GetUserBoards");
            return BoardsFactory.Instance.GetUserBoards(user.Id);
        }

        //Retorna la llista de taulers disponibles (els del usuari mes els que hi hagi per defecte)
        public List<BoardDTO> GetBoards()
        {
            Console.WriteLine(user.Name + ": GetBoards");
            return BoardsFactory.Instance.GetBoards(user.Id);
        }

        //Retorna la matriu de caselles d'un tauler determinat
        public List<List<BoxDTO>> GetBoardBoxes(string boardId)
        {
            Console.WriteLine(user.Name + ": GetBoardBoxes: " + boardId);
            return BoardsFactory.Instance.GetBoardBoxes(boardId);
        }

        //Desa un tauler a la base de dades
        public void SaveBoard(BoardDTO board, List<List<BoxDTO>> boxes)
        {
            Console.WriteLine(user.Name + ": SaveBoard: " + board.Id + " -- " + board.Name);
            BoardsFactory.Instance.SaveBoard(board, boxes, user.Id);
        }

        //Elimina un tauler de la base de dades
        public void DeleteBoard(BoardDTO board)
        {
            Console.WriteLine(user.Name + ": DeleteBoard: " + board.Id + " -- " + board.Name);
            BoardsFactory.Instance.DeleteBoard(board, user.Id);
        }

        /*
         * *************************************************************************************
         * GESTIÓ DE SACS DE FITXES
         * *************************************************************************************
         */

        //Retorna la llista de sacs de fitxes de l'usuari identificat
        public List<TileSetDTO> GetUserTileSets()
        {
            Console.WriteLine(user.Name + ": GetUserTileSets");
            return TileSetsFactory.Instance.GetUserTileSets(user.Id);
        }

        //Retorna la llista de sacs de fitxesdisponibles (els del usuari mes els que hi hagi per defecte)
        public List<TileSetDTO> GetTileSets()
        {
            Console.WriteLine(user.Name + ": GetTileSets");
            return TileSetsFactory.Instance.GetTileSets(user.Id);
        }

        //Retorna la llista de fitxes d'un sac de fitxes determinat
        public List<TileDTO> GetTileSetTiles(string tileSetId)
        {
            Console.WriteLine(user.Name + ": GetTileSetTiles: " + tileSetId);
            return TileSetsFactory.Instance.GetTileSetTiles(tileSetId);
        }

        //Desa un sac de fitxes a la base de dades
        public void SaveTileSet(TileSetDTO tileSet, List<TileDTO> tiles)
        {
            Console.WriteLine(user.Name + ": SaveTileSet: " + tileSet.Id + " -- " + tileSet.Name);
            TileSetsFactory.Instance.SaveTileSet(tileSet, tiles, user.Id);
        }

        //Elimina un sac de fitxes dela base de dades
        public void DeleteTileSet(TileSetDTO tileSet)
        {
            Console.WriteLine(user.Name + ": DeleteTileSet: " + tileSet.Id + " -- " + tileSet.Name);
            TileSetsFactory.Instance.DeleteTileSet(tileSet, user.Id);
        }


        /*
        * *************************************************************************************
        * GESTIÓ DE DICCIONARIS
        * *************************************************************************************
        */

        //Retorna la llista de informació de diccionaris
        public List<DictionaryInfoDTO> GetDictionaryInfos()
        {
            Console.WriteLine(user.Name + ": GetDictionaryInfos");

            return DictionariesFactory.Instance.GetDictionaryInfos();
        }

        //Desa una informació de diccionari a la base de dades (gestió nomes per usuaris administradors)
        public void SaveDictionaryInfo(DictionaryInfoDTO dictionaryInfo)
        {
            Console.WriteLine(user.Name + ": SaveDictionaryInfo");

            EnsureUserIsAdmin();
            DictionariesFactory.Instance.SaveDictionaryInfo(dictionaryInfo);
        }

        //Elimina una informació de diccionari de la base de dades (gestió nomes per usuaris administradors)
        public void DeleteDictionaryInfo(DictionaryInfoDTO dictionaryInfo)
        {
            Console.WriteLine(user.Name + ": DeleteDictionaryInfo");

            EnsureUserIsAdmin();
            DictionariesFactory.Instance.DeleteDictionaryInfo(dictionaryInfo);
        }



        /*
         * *************************************************************************************
         * GESTIÓ DE PARTIDES
         * *************************************************************************************
         */

        //Crea una nova partida per un sol jugador
        public GameInfoDTO NewOnePlayerGame(GameInfoDTO gameInfo)
        {
            Console.WriteLine(user.Name + ": NewOnePlayerGame");

            return GameManager.Instance.NewOnePlayerGame(gameInfo, user);
        }

        //Crea una nova partida per dos jugadors
        public GameInfoDTO NewTwoPlayersGame(GameInfoDTO gameInfo)
        {
            Console.WriteLine(user.Name + ": NewTwoPlayersGame");

            return GameManager.Instance.NewTwoPlayersGame(gameInfo, user);
        }

        //Elimina una partida
        public void DeleteGame(string gameId)
        {
            Console.WriteLine(user.Name + ": DeleteGame");

            GameManager.Instance.DeleteGame(gameId, user);
        }

        //Busca les partides pendents d'afegir un segon jugador
        public List<GameInfoDTO> GetAvailableGames()
        {
            Console.WriteLine(user.Name + ": GetAvailableGames");

            return GameManager.Instance.GetAvailableGames(user);
        }

        //Busca les partides en que participa l'usuari
        public List<GameInfoDTO> GetCurrentUserGames()
        {
            Console.WriteLine(user.Name + ": GetCurrentUserGames");

            return GameManager.Instance.GetCurrentUserGames(user);
        }

        //Afegeix a l'usuari com a segon jugador d'una partida
        public GameInfoDTO AddMeAsSecondPlayer(string gameId)
        {
            Console.WriteLine(user.Name + ": AddMeAsSecondPlayer");

            return GameManager.Instance.AddSecondPlayer(gameId, user);
        }

        //Retorna la informació bàsica d'una partida actualitzada
        public GameInfoDTO GetGameInfo(string gameId)
        {
            Console.WriteLine(user.Name + ": GetGameInfo");

            return GameManager.Instance.GetGameInfo(gameId);
        }

        //Retorna l'estat del tauler en un moment determinat
        public List<List<BoxDTO>> GetGameBoardStatus(string gameId)
        {
            Console.WriteLine(user.Name + ": GetGameBoardStatus");

            return GameManager.Instance.GetGameBoardStatus(gameId);
        }

        //Retorna el faristol d'una partida del jugador
        public List<TileDTO> GetGameRack(string gameId)
        {
            Console.WriteLine(user.Name + ": GetGameRack");

            return GameManager.Instance.GetGameRack(gameId, user);
        }

        //Senyal indicant que es pot iniciar la partida
        public void StartGame(string gameId)
        {
            Console.WriteLine(user.Name + ": StartGame");

            GameManager.Instance.StartGame(gameId, user);
        }

        //Indica que l'usuari s'està adjuntant a la partida (retorna si es el torn o no de l'usuari)
        public bool AttachToGame(string gameId)
        {
            Console.WriteLine(user.Name + ": AttachToGame");

            return GameManager.Instance.AttachToGame(gameId, user);
        }

        //Renuncia al torn i el passa l'altre jugador
        public void PassTurn(string gameId)
        {
            Console.WriteLine(user.Name + ": PassTurn");

            GameManager.Instance.PassTurn(gameId, user);
        }

        //Canvia les fitxes per unes altres del sac
        public List<TileDTO> SwapTiles(string gameId, List<TileDTO> tiles)
        {
            Console.WriteLine(user.Name + ": SwapTiles");

            return GameManager.Instance.SwapTiles(gameId, user, tiles);
        }

        //Retorna la llista de lletres vàlides del sac de fitxes associat a una partida
        public List<string> GetValidLetters(string gameId)
        {
            Console.WriteLine(user.Name + ": GetValidLetters");

            return GameManager.Instance.GetValidLetters(gameId, user);
        }


        //L'usuari es rendeix i dona com a guanyador el rival
        public void Resign(string gameId)
        {
            Console.WriteLine(user.Name + ": Resign");

            GameManager.Instance.Resign(gameId, user);
        }


        //Comprova els punts que dona una jugada
        public MoveInfoDTO CheckMove(MoveDTO move)
        {
            Console.WriteLine(user.Name + ": CheckMove");

            return GameManager.Instance.CheckMove(move, user);
        }

        //Realitza una jugada
        public MoveInfoDTO MakeMove(MoveDTO move)
        {
            Console.WriteLine(user.Name + ": MakeMove");

            return GameManager.Instance.MakeMove(move, user);
        }


        /*
         * *************************************************************************************
         * CONSULTA D'ESTADÍSTIQUES
         * *************************************************************************************
         */

        //Torna el Top10 de de millors partides
        public List<StatsHighGameDTO> GetHighestGameScores(bool onlyUserStats, bool excludeCPU)
        {
            Console.WriteLine(user.Name + ": GetHighestGameScores");

            return GamesFactory.Instance.GetHighestGameScores(onlyUserStats, excludeCPU, user.Id);
        }

        //Torna el Top10 de de millors jugades
        public List<StatsHighMoveDTO> GetHighestMoveScores(bool onlyUserStats, bool excludeCPU)
        {
            Console.WriteLine(user.Name + ": GetHighestMoveScores");

            return GamesFactory.Instance.GetHighestMoveScores(onlyUserStats, excludeCPU, user.Id);
        }

        //Torna el Top10 de de millors percentatges de partides guanyades
        public List<StatsGamePercentages> GetHighestWinGamesPercent(bool onlyUserStats, bool excludeCPU)
        {
            Console.WriteLine(user.Name + ": GetHighestWinGamesPercent");

            return GamesFactory.Instance.GetHighestWinGamesPercent(onlyUserStats, excludeCPU, user.Id);
        }

        //Elimina les estadístiques de l'usuari
        public void DeleteUserStats()
        {
            Console.WriteLine(user.Name + ": DeleteUserStats");

            GamesFactory.Instance.DeleteUserStats(user.Id);
        }
    }
}
